"""SNAPstraction Serial/Flash ops"""
################################################################
# Copied from flasher.FlashUpgrader, mostly. Adapted for CLI and futures
import array
import binascii
import bz2
import functools
import importlib
import logging
import os
import tempfile

from flasher.Flasher import Hcs08Flasher
from flasher.ATmega128RFA1Flasher import ATMegaFlasher

from apy import EventScheduler

from tornado.ioloop import IOLoop, PeriodicCallback, TimeoutError
from tornado.concurrent import Future
from tornado.gen import coroutine, Return

import serialwrapper

from util import ensure_path_contains_python_lib

log = logging.getLogger(__name__)

RF100_MODULES = ['RF100']


class FlasherError(Exception):
    pass


def get_flasher_for_module(module, *args, **kwargs):
    """Returns the flasher class for the module type"""
    if module in RF100_MODULES:
        return Hcs08Flasher(*args, **kwargs)
    else:
        return ATMegaFlasher(*args, **kwargs)


def _get_file_path_for_version(metadata, module, version=None):
    img_dir = os.path.join(metadata._get_meta_dir(), '..', 'images')
    filename = metadata.get_metadata(module, version, aes=True)['filename']
    return os.path.join(img_dir, filename)


def firmware(profile, module, core, sniffer=False, override_file=None, timeout=180, progress_cb=None, finished_cb=None):
    """Flash firmware"""
    firmware = None
    if override_file:
        firmware = os.path.abspath(os.path.expanduser(override_file))
        log.warn("Overriding firmware path: %s", firmware)
    else:
        # Extend the path so we can find firmware support packages
        ensure_path_contains_python_lib()

        if sniffer:
            metadata_module_name = "snap_sniffer"
            log.debug("Looking for sniffer firmware for %s", metadata_module_name)
        else:
            metadata_module_name = "snap_firmware_%s" % (core.replace('.', '_'),)
            log.debug("Looking for firmware for %s", metadata_module_name)

        try:
            metadata = importlib.import_module(metadata_module_name + '.metadata')
        except:
            log.error(
                "Could not find snap-firmware-%s. Run: pip install --extra-index-url https://update.synapse-wireless.com/pypi snap-firmware-%s",
                core, core)
            raise

        try:
            firmware = _get_file_path_for_version(metadata, module, core)
        except:
            log.error("Module %s is not supported by firmware %s.", module, core)
            raise

        log.debug("Found firmware %s", firmware)

    # Convert SFI to HEX...
    if os.path.splitext(firmware)[1] == '.sfi':
        fd, path = tempfile.mkstemp(text=True)
        sfi_contents = open(firmware, 'rb').read()
        decompressText = bz2.decompress(sfi_contents)
        if decompressText is not None:
            tmpFile = os.fdopen(fd, 'w')
            tmpFile.write(decompressText)
            tmpFile.close()
            firmware = path

    cb = functools.partial(flasher, profile, module, firmware, "Flash Firmware", progress_cb=progress_cb, finished_cb=finished_cb)
    flash_result = None

    try:
        flash_result = IOLoop.current().run_sync(cb, timeout=timeout)
    except TimeoutError:
        log.error("Flash firmware timed out or aborted!")

    return flash_result


def default_nv(profile, module, timeout=30, progress_cb=None, finished_cb=None):
    """Factory default the NV params"""
    fd, path = tempfile.mkstemp(text=True)
    tempFile = os.fdopen(fd, 'w')

    if module in RF100_MODULES:
        tempFile.write(buildMagicSRec('N'))
    else:
        tempFile.write(buildMagicHRec('N'))

    tempFile.close()

    cb = functools.partial(flasher, profile, module, path, "Default NV Params", progress_cb=progress_cb, finished_cb=finished_cb)
    flash_result = None
    try:
        flash_result = IOLoop.current().run_sync(cb, timeout=timeout)
    except TimeoutError:
        log.error("Default NV Params timed out or aborted!")

    return flash_result


def erase_script(profile, module, timeout=30, progress_cb=None, finished_cb=None):
    """Erase the SNAPpy script"""
    fd, path = tempfile.mkstemp(text=True)
    tempFile = os.fdopen(fd, 'w')

    if module in RF100_MODULES:
        tempFile.write(buildMagicSRec('S'))
    else:
        tempFile.write(buildMagicHRec('S'))

    tempFile.close()

    cb = functools.partial(flasher, profile, module, path, "Erase SNAPpy Script", progress_cb=progress_cb, finished_cb=finished_cb)
    flash_result = None
    try:
        flash_result = IOLoop.current().run_sync(cb, timeout=timeout)
    except TimeoutError:
        log.error("Erase Script timed out or aborted!")

    return flash_result


def e20_reset_on():
    """Toggle the E20 radio reset pin low"""
    try:
        with open('/sys/class/gpio/gpio34/value', 'w') as gpio:
            gpio.write('0')
    except:
        log.exception("Couldn't reset SNAP node! (Do you need sudo?)")
        raise


def e20_reset_off():
    """Toggle the E20 radio reset pin back to high"""
    try:
        with open('/sys/class/gpio/gpio34/value', 'w') as gpio:
            gpio.write('1')
    except:
        log.exception("Couldn't reset SNAP node! (Do you need sudo?)")
        raise


def e12_reset_on():
    """Toggle the E12 radio reset pin low"""
    try:
        with open('/sys/class/gpio/gpio48/value', 'w') as gpio:
            gpio.write('0')
    except:
        log.exception("Couldn't reset SNAP node! (Do you need sudo?)")
        raise


def e12_reset_off():
    """Toggle the E12 radio reset pin back to high"""
    try:
        with open('/sys/class/gpio/gpio48/value', 'w') as gpio:
            gpio.write('1')
    except:
        log.exception("Couldn't reset SNAP node! (Do you need sudo?)")
        raise


def bitbang_SS200_reset_on(serial):
    """Bitbang the SS200 reset line low"""
    serial.setBitMode(0x10, 0x20)


def bitbang_SS200_reset_off(serial):
    """Bitbang the SS200 reset line back to high"""
    serial.setBitMode(0x11, 0x20)


def try_setDTR(serial):
    """Try to set DTR"""
    try:
        log.debug("setting DTR")
        # Yep, this is inverted on purpose
        serial.setDTR(False)
    except:
        log.exception("Couldn't setDTR")


def try_clearDTR(serial):
    """Try to clear DTR"""
    try:
        log.debug("clearing DTR")
        # Yep, this is inverted on purpose
        serial.setDTR(True)
    except:
        log.exception("Couldn't setDTR")


def get_radio_reset_funcs(serial_type, serial_port, serial_inst):
    """Returns functions and a suggested delay for resetting a SNAP radio.

    Args:
        serial_type (int): SNAPconnect Serial Port Type (see SNAPconnect documentation)
        serial_port (int|string): SNAPconnect Serial Port (e.g. 0, "/dev/snap1", "/dev/ttyUSB0",
            see SNAPconnect documentation for more details)
        serial_inst (serialwrapper.serial.serialutil.SerialBase): PySerial instance w/ setBitMode and setDTR methods

    Returns:
        Tuple of (reset_on_func, reset_off_func, delay_in_seconds)

    """
    if serial_type == serialwrapper.PyserialDriver.PyserialWrapper.TYPE_SNAPSTICK:
        log.info("Using an SS200, resetting via bitbanged GPIO")
        return functools.partial(bitbang_SS200_reset_on, serial_inst), functools.partial(bitbang_SS200_reset_off, serial_inst), 2

    elif os.path.exists('/etc/e20_version') and serial_port == '/dev/snap1':
        log.info("Using /dev/snap1 on an E20, we'll attempt to reset via GPIO 34")
        return e20_reset_on, e20_reset_off, 1

    elif os.path.exists('/etc/e12_version') and serial_port == '/dev/snap1':
        log.info("Using /dev/snap1 on an E12, we'll attempt to reset via GPIO 48")
        return e12_reset_on, e12_reset_off, 1

    else:
        return functools.partial(try_setDTR, serial_inst), functools.partial(try_clearDTR, serial_inst), 1


@coroutine  # noqa
def flasher(profile, module, path, action, progress_cb=None, finished_cb=None):
    """Workhorse, actually runs the flasher"""
    result = Future()

    evScheduler = EventScheduler.EventScheduler()

    serialwrapper_path = os.path.dirname(serialwrapper.__file__)

    def on_finished():
        """Complete callback"""
        log.info("%s complete!", action)
        if finished_cb:
            finished_cb()
        result.set_result(True)

    sc_profile = profile.as_sc_params()

    f = get_flasher_for_module(module,
                               path,
                               scheduler=evScheduler,
                               finishedCallback=on_finished,
                               port=sc_profile.port,
                               type=sc_profile.type,
                               pathToUsbLibrary=serialwrapper_path)

    log.debug("Flasher Instance is: %r", f)
    log.debug("Flasher Vars: %r", vars(f))

    if module not in RF100_MODULES:
        # Force the flasher to idle
        f.state = f.STATE_IDLE

        def go_flasher():
            """Tell the flasher it can wait for incoming data now"""
            f.state = f.STATE_INCOMING_WAIT
    else:
        # RF100 / Hcs08Flasher doesn't have the same state machine
        def go_flasher():
            """Don't do anything"""
            pass

    def on_progress(counter):
        """Progress update callback"""
        log.debug("Got flasher update: %r", counter)
        if progress_cb:
            progress_cb(counter, f.max_progress)

    def on_error(error):
        """Error callback"""
        log.error("Flasher error: %r", error)
        result.set_exception(FlasherError(error))

    f.registerErrorCallback(on_error)
    f.registerProgressCallback(on_progress)

    log.warn("Restart your device.")
    if module in RF100_MODULES:
        log.warn("RF100s currently don't reset-on-DTR, you must reset them manually.")
    else:
        log.warn("We'll try to restart it for you if it's an SS200 / reset-on-DTR.")

    reset_on, reset_off, delay = get_radio_reset_funcs(sc_profile.type, sc_profile.port, f.serialDrv.serial)

    evScheduler.schedule(delay + 0.0, reset_on)
    evScheduler.schedule(delay + 0.1, go_flasher)
    evScheduler.schedule(delay + 0.2, reset_off)

    PeriodicCallback(evScheduler.poll, 5).start()

    done = yield result
    raise Return(done)

################################
# These are copied from flasher, because the file they're in now also imports wx...
# Favor duplication over dealing with wxPython...


def s19_checksum(line):
    """Calculate the S19 checksum"""
    length = int(line[2:4], 16)
    pbytes = [int(line[i * 2:i * 2 + 2], 16) for i in range(1, length + 1)]
    return ~sum(pbytes) & 0xff


def niceHex(number):
    """Format a number in a nice hex way?"""
    numHex = hex(number)
    numHex = numHex[2:]  # chop off the leading '0x'
    if len(numHex) != 2:
        numHex = '0' + numHex  # tack on a leading '0' if not two digits

    numHex = numHex.upper()
    return numHex


def buildMagicSRec(cmdChar):
    """Build the magic S19 record used for defaulting NV params and erasing scripts"""
    magicAddr = '1080'
    magicKey = '524D44454D474B42'

    cmdHex = niceHex(ord(cmdChar))

    magicLen = niceHex((len(magicAddr) + len(magicKey) + len(cmdHex) + 2) / 2)  # +2 for trailing checksum, /2 to get bytes not chars

    srec = 'S1' + magicLen + magicAddr + magicKey + cmdHex

    checksum = s19_checksum(srec)
    srec = srec + niceHex(checksum)

    return srec


def buildMagicHRec(cmdChar, addr='02F0'):
    """Build the magic hex record used for defaulting NV params and erasing scripts"""
    magicKey = '524D44454D474B42'
    record = '09%s00%s%02X' % (addr, magicKey, ord(cmdChar))
    crc = (~sum(array.array('B', binascii.unhexlify(record))) + 1) % 2 ** 8

    # text = ':020000020000FC\n' # Header
    text = ':%s%02X\n' % (record, crc)
    text += ':00000001FF'  # End of file record

    return text
