import logging
import mock
import unittest
import importlib
import tempfile
import os

import snap_profile

from tests import helpers

from tornado.testing import AsyncTestCase, gen_test

from snapstraction import flash


log = logging.getLogger(__name__)


class MockAtmelFlasher(object):
    STATE_IDLE = 0
    STATE_INCOMING_WAIT = 1

    def __init__(self, path, scheduler=None, finishedCallback=None, port=None, type=None, pathToUsbLibrary=None):
        log.info("Creating Mock Atmel Flasher: %r, %r", path, port)
        self.update_callback = None
        self.error_callback = None
        self.complete_callback = finishedCallback
        self.max_progress = 100
        self.state = self.STATE_INCOMING_WAIT
        self.serialDrv = mock.MagicMock()
        self.serialDrv.serial = mock.MagicMock()

    def start(self, probeVer, autoRunCallback):
        self.probeVer = probeVer
        self.complete_callback = autoRunCallback

    def registerErrorCallback(self, errorCallback):
        self.error_callback = errorCallback

    def registerProgressCallback(self, updateCallback):
        self.update_callback = updateCallback

    def mock_error(self, error):
        if self.error_callback:
            self.error_callback(error)

    def mock_update(self, result):
        if self.update_callback:
            self.update_callback(result)

    def mock_complete(self):
        if self.complete_callback:
            self.complete_callback()


class FlashCommandTests(unittest.TestCase):

    def setUp(self):
        # Make our tests operate on a temp file
        fd, path = tempfile.mkstemp()
        self.path = path
        snap_profile.profile._env_override = mock.MagicMock(return_value=path)

        with open(path, 'wt') as f:
            f.write("""
[default]
device = /dev/snap1
kind = serial
address =
""")
        super(FlashCommandTests, self).setUp()

    @mock.patch.object(flash, 'ATMegaFlasher')
    def test_get_flasher_for_module_Atmels(self, mock_flasher):
        mock_flasher.return_value = "Fake!"
        f = flash.get_flasher_for_module("SN220", "mock.hex")
        self.assertEqual(f, "Fake!")

    @mock.patch.object(flash, 'Hcs08Flasher')
    def test_get_flasher_for_module_RFEngine(self, mock_flasher):
        mock_flasher.return_value = "Fake!"
        f = flash.get_flasher_for_module("RF100", "mock.hex")
        self.assertEqual(f, "Fake!")

    @mock.patch.object(flash, 'flasher')
    def test_default_nv_params_calls_flasher(self, cmd):
        cmd.return_value = helpers.make_resolved_future(True)

        profile = snap_profile.profile.Profile()

        flash.default_nv(profile, 'SN220')

        cmd.assert_called_with(profile, 'SN220', mock.ANY, "Default NV Params", progress_cb=None, finished_cb=None)

    @mock.patch.object(flash, 'flasher')
    def test_erase_script_calls_flasher(self, cmd):
        cmd.return_value = helpers.make_resolved_future(True)

        profile = snap_profile.profile.Profile()

        flash.erase_script(profile, 'SN220')

        cmd.assert_called_with(profile, 'SN220', mock.ANY, "Erase SNAPpy Script", progress_cb=None, finished_cb=None)

    @mock.patch.object(flash, 'flasher')
    def test_firmware_load_calls_flasher(self, cmd):
        cmd.return_value = helpers.make_resolved_future(True)

        profile = snap_profile.profile.Profile()

        flash.firmware(profile, 'SN220', '2.6.9', sniffer=False, override_file="override.hex")

        cmd.assert_called_with(profile, 'SN220', mock.ANY, "Flash Firmware", progress_cb=None, finished_cb=None)

    @mock.patch.object(importlib, 'import_module')
    @mock.patch.object(flash, 'flasher')
    def test_firmware_load_with_lookup(self, cmd, importer):
        mock_sdk = mock.MagicMock()
        mock_sdk.SNYF_FIRMWARE = {'1.3.0': 'mock.hex'}
        mock_sdk.CORE_FIRMWARE = {'2.6.9': 'mock.hex'}

        importer.return_value = mock_sdk
        cmd.return_value = helpers.make_resolved_future(True)

        profile = snap_profile.profile.Profile()

        flash.firmware(profile, 'SN220', '2.6.9', sniffer=False)

        cmd.assert_called_with(profile, 'SN220', mock.ANY, "Flash Firmware", progress_cb=None, finished_cb=None)


class FlashHelperTests(unittest.TestCase):

    def test_s19_checksum(self):
        line = "S123182C0000BA708B899EFE05F6AF019EFF05888A81A7FCC618C34C95E701C618C24CF7EE0"
        checksum = flash.s19_checksum(line)

        self.assertEqual(checksum, 238)

    def test_niceHex(self):
        for i in xrange(0xff):
            self.assertEqual(flash.niceHex(i), ("%02x" % i).upper())

        for i in xrange(0xf):
            n = 0x100 + i
            self.assertEqual(flash.niceHex(n), ("%04x" % n).upper())

    def test_buildMagicSRec(self):
        self.assertEqual(flash.buildMagicSRec('S'), 'S10C1080524D44454D474B4253C7')
        self.assertEqual(flash.buildMagicSRec('N'), 'S10C1080524D44454D474B424ECC')

    def test_buildMagicHRec(self):
        self.assertEqual(flash.buildMagicHRec('S'), ':0902F000524D44454D474B425369\n:00000001FF')
        self.assertEqual(flash.buildMagicHRec('N'), ':0902F000524D44454D474B424E6E\n:00000001FF')


class FlasherTests(AsyncTestCase):

    def setUp(self):
        # Make our tests operate on a temp file
        fd, path = tempfile.mkstemp()
        self.path = path
        snap_profile.profile._env_override = mock.MagicMock(return_value=path)

        with open(path, 'wt') as f:
            f.write("""
[default]
device = /dev/snap1
kind = serial
address =
""")
        self.mock_flasher = MockAtmelFlasher("mock.hex")

        def get_mock_flasher_for_module(module, path, scheduler, finishedCallback, **kwargs):
            self.mock_flasher.complete_callback = finishedCallback
            return self.mock_flasher

        flash.get_flasher_for_module = get_mock_flasher_for_module

        self.profile = snap_profile.profile.Profile.load('default')

        super(FlasherTests, self).setUp()

    @gen_test
    def test_flasher_calls_flasher(self):
        f = flash.flasher(self.profile, 'SN220', "override.hex", "Mock Flash")
        self.mock_flasher.mock_complete()
        results = yield f
        self.assertEqual(results, True)

    @gen_test
    def test_flasher_can_update(self):
        f = flash.flasher(self.profile, 'SN220', "override.hex", "Mock Flash")
        self.mock_flasher.mock_update(10)
        self.mock_flasher.mock_update(20)
        self.mock_flasher.mock_update(30)
        self.mock_flasher.mock_complete()
        results = yield f
        self.assertEqual(results, True)

    @gen_test
    def test_flasher_dies_on_error(self):
        f = flash.flasher(self.profile, 'SN220', "override.hex", "Mock Flash")
        self.mock_flasher.mock_update(10)
        self.mock_flasher.mock_error("Something happened")
        with self.assertRaises(flash.FlasherError):
            yield f


class RadioResetTests(unittest.TestCase):

    def setUp(self):
        self.mock_serial = mock.MagicMock()

    def test_default_is_DTR_at_1_second(self):
        reset_on, reset_off, delay = flash.get_radio_reset_funcs(snap_profile.profile.SERIAL_TYPE_RS232, 0, self.mock_serial)

        reset_on()
        self.mock_serial.setDTR.assert_called_once_with(False)

        self.mock_serial.reset_mock()

        reset_off()
        self.mock_serial.setDTR.assert_called_once_with(True)

        self.assertEqual(delay, 1)

    def test_SS200_is_BitBang_at_2_seconds(self):
        reset_on, reset_off, delay = flash.get_radio_reset_funcs(snap_profile.profile.SERIAL_TYPE_SNAPSTICK200, 0, self.mock_serial)

        reset_on()
        self.mock_serial.setBitMode.assert_called_once_with(0x10, 0x20)

        self.mock_serial.reset_mock()

        reset_off()
        self.mock_serial.setBitMode.assert_called_once_with(0x11, 0x20)

        self.assertEqual(delay, 2)

    def test_E20_snap1_is_GPIO34_at_1_second(self):

        def mock_e20_exists(path):
            return path == "/etc/e20_version"

        with mock.patch("os.path.exists", new=mock_e20_exists):
            reset_on, reset_off, delay = flash.get_radio_reset_funcs(snap_profile.profile.SERIAL_TYPE_RS232, "/dev/snap1", self.mock_serial)

            mock_open = mock.mock_open()
            with mock.patch("snapstraction.flash.open", mock_open, create=True):
                reset_on()
                mock_open.assert_called_once_with("/sys/class/gpio/gpio34/value", "w")
                handle = mock_open()
                handle.write.assert_called_once_with("0")

                mock_open.reset_mock()

                reset_off()
                mock_open.assert_called_once_with("/sys/class/gpio/gpio34/value", "w")
                handle = mock_open()
                handle.write.assert_called_once_with("1")

            self.assertEqual(delay, 1)

    def test_E12_snap1_is_GPIO48_at_1_second(self):

        def mock_e12_exists(path):
            return path == "/etc/e12_version"

        with mock.patch("os.path.exists", new=mock_e12_exists):
            reset_on, reset_off, delay = flash.get_radio_reset_funcs(snap_profile.profile.SERIAL_TYPE_RS232, "/dev/snap1", self.mock_serial)

            mock_open = mock.mock_open()
            with mock.patch("snapstraction.flash.open", mock_open, create=True):
                reset_on()
                mock_open.assert_called_once_with("/sys/class/gpio/gpio48/value", "w")
                handle = mock_open()
                handle.write.assert_called_once_with("0")

                mock_open.reset_mock()

                reset_off()
                mock_open.assert_called_once_with("/sys/class/gpio/gpio48/value", "w")
                handle = mock_open()
                handle.write.assert_called_once_with("1")

            self.assertEqual(delay, 1)

    def test_E20_SN220_is_DTR_at_1_second(self):

        def mock_e20_exists(path):
            return path == "/etc/e20_version"

        with mock.patch("os.path.exists", new=mock_e20_exists):
            reset_on, reset_off, delay = flash.get_radio_reset_funcs(snap_profile.profile.SERIAL_TYPE_RS232, "/dev/ttyUSB0", self.mock_serial)

            reset_on()
            self.mock_serial.setDTR.assert_called_once_with(False)

            self.mock_serial.reset_mock()

            reset_off()
            self.mock_serial.setDTR.assert_called_once_with(True)

            self.assertEqual(delay, 1)

    def test_E12_SN220_is_DTR_at_1_second(self):

        def mock_e12_exists(path):
            return path == "/etc/e12_version"

        with mock.patch("os.path.exists", new=mock_e12_exists):
            reset_on, reset_off, delay = flash.get_radio_reset_funcs(snap_profile.profile.SERIAL_TYPE_RS232, "/dev/ttyUSB0", self.mock_serial)

            reset_on()
            self.mock_serial.setDTR.assert_called_once_with(False)

            self.mock_serial.reset_mock()

            reset_off()
            self.mock_serial.setDTR.assert_called_once_with(True)

            self.assertEqual(delay, 1)
