"""Tests for the Grand Unified Snap interface.

Assumes: SNAP Connect, SNAP Connect Futures, and SNAP2MQTT client libraries work."""
import logging

import mock
from mock import MagicMock, patch, call, ANY
import unittest

from timeit import default_timer as timer

from datetime import timedelta

from snap_profile.profile import SerialProfile, TcpProfile, MqttProfile
from snap_profile.network import Network

# Mock out the underlying bits for this entire module
# patch('snapconnect.snap').start()
patch('snapconnect_futures.SnapConnectFutures').start()
patch('snap2mqtt_client.client.BaseClient').start()
patch('snap2mqtt_future_client.future_client.FutureClient').start()

import snapstraction.gus

from snapstraction.gus import Gus, SCFGus, GusError, NoResponseError, coerce_priority

from snapconnect_futures.responses import MultiResult

from snappy import SnapConstantsHelper as SnapConst

from tornado.concurrent import Future
from tornado.gen import coroutine, with_timeout, sleep, Return
from tornado.testing import AsyncTestCase, gen_test

from tests import helpers

from snapstraction.mocks.arbiter import Arbiter

log = logging.getLogger(__name__)

snap_mock = None
scf_mock = None
snap_patch = patch('snapstraction.gus.snap.Snap')
scf_patch = patch('snapstraction.gus.SnapConnectFutures')


def setUpModule():
    global snap_mock
    global scf_mock
    snap_mock = snap_patch.start()
    scf_mock = scf_patch.start()


def tearDownModule():
    snap_patch.stop()
    scf_patch.stop()


def build_gus(profile, network, license):
    sc = MagicMock()
    snap_mock.return_value = sc
    sc.licensed_mac = '\x12\x34\x56'
    sc.save_nv_param = MagicMock()

    scf = MagicMock()
    scf_mock.return_value = scf
    scf.rpc.return_value = helpers.make_resolved_future((None,))
    scf.mcast_rpc.return_value = helpers.make_resolved_future((None,))
    scf.dmcast_rpc.return_value = helpers.make_resolved_future((None,))

    scf.callback_rpc.return_value = helpers.make_resolved_future((None,))
    mcast_result = MultiResult({}, {})
    mcast_result.retry = MagicMock()
    mcast_result.retry.return_value = helpers.make_resolved_future(None)
    scf.callback_mcast_rpc.return_value = helpers.make_resolved_future(mcast_result)
    dmcast_result = MultiResult({'123456': (True,)}, {})
    dmcast_result.retry = MagicMock()
    dmcast_result.retry.return_value = helpers.make_resolved_future(None)
    scf.callback_dmcast_rpc.return_value = helpers.make_resolved_future(dmcast_result)

    scf._snap = sc

    gus = Gus(profile, network, license)
    gus._schedule_tx_service()

    return gus


class TestGus(unittest.TestCase):
    """Tests for creating Gus instances"""

    def test_gus_no_profile_raises(self):
        with self.assertRaises(Exception):
            Gus(None)

    def test_gus_bad_profile_raises(self):
        with self.assertRaises(GusError):
            class Object(object):
                pass
            bad_profile = Object()
            bad_profile.kind = "LOLWUT?"
            network = Network()
            Gus(bad_profile, network)


class TestSCFGusFromExisting(unittest.TestCase):
    """Tests for SFCGus from existing instance"""

    def test_gus_from_existing_calls_SCFGus(self):
        scf_instance = MagicMock()
        SCFGus_from_existing = Gus.from_existing_scf(scf_instance, bridge_addr=123456)

        self.assertIsInstance(SCFGus_from_existing, SCFGus)
        self.assertEquals(SCFGus_from_existing._scf, scf_instance)
        self.assertEquals(SCFGus_from_existing.bridge_addr, 123456)

    def test_SCFGus_from_existing_without_bridge_addr(self):
        scf_instance = MagicMock()
        SCFGus_from_existing = Gus.from_existing_scf(scf_instance)

        self.assertIsInstance(SCFGus_from_existing, SCFGus)
        self.assertEquals(SCFGus_from_existing._scf, scf_instance)
        self.assertEquals(SCFGus_from_existing.bridge_addr, None)

    def test_SCFGus_from_existing_doesnt_overwrite_existing_values(self):
        scf_instance = MagicMock()
        SCFGus_from_existing = Gus.from_existing_scf(scf_instance)

        self.assertIsInstance(SCFGus_from_existing, SCFGus)
        self.assertEquals(SCFGus_from_existing._sc, scf_instance._snap)
        self.assertFalse(hasattr(SCFGus_from_existing, 'profile'))
        self.assertFalse(hasattr(SCFGus_from_existing, 'network'))

    def test_SCFGus_from_existing_start_doesnt_call_init(self):
        scf_instance = MagicMock()
        SCFGus_from_existing = Gus.from_existing_scf(scf_instance)

        SCFGus_from_existing._init_for_serial = MagicMock()
        SCFGus_from_existing._init_for_tcp = MagicMock()

        SCFGus_from_existing.start()

        SCFGus_from_existing._init_for_serial.assert_not_called()
        SCFGus_from_existing._init_for_tcp.assert_not_called()

    def test_SCFGus_from_existing_stop_doesnt_call_close(self):
        scf_instance = MagicMock()
        SCFGus_from_existing = Gus.from_existing_scf(scf_instance)

        SCFGus_from_existing._close_for_serial = MagicMock()
        SCFGus_from_existing._close_for_tcp = MagicMock()

        SCFGus_from_existing.stop()

        SCFGus_from_existing._close_for_serial.assert_not_called()
        SCFGus_from_existing._close_for_tcp.assert_not_called()


class TestMQTTGus(AsyncTestCase):
    """Tests for MQTT Gus wrapper"""

    def get_gus(self):
        profile = MqttProfile()
        network = Network()
        return Gus(profile, network)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_init_good_profile(self, base_client, future_client):
        gus = self.get_gus()
        self.assertIsNotNone(gus)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_context_manager(self, base_client, future_client):
        gus = self.get_gus()
        gus.start = MagicMock()
        gus.stop = MagicMock()

        gus.__enter__()
        gus.__exit__(None, None, None)

        gus.start.assert_called_once_with()
        gus.stop.assert_called_once_with()

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_implement_rpc_funcs(self, base_client, future_client):
        try:
            gus = self.get_gus()

            gus.ucast_rpc('123456', 'ping')
            gus.mcast_rpc(1, 5, 'ping')
            gus.dmcast_rpc(['123456', '234567'], 1, 5, 50, 'ping')

            gus.callback_ucast_rpc('123456', 'ping')
            gus.callback_mcast_rpc(1, 5, 'ping')
            gus.callback_dmcast_rpc(['123456', '234567'], 1, 5, 50, 'ping')

        except:
            log.exception("Unexpected exception")
            self.fail("Unexpected exception")

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_ucast_rpc(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.ucast_rpc('123456', 'ping', 123, 'abc')
        fc.uRPC.assert_called_once_with('123456', 'ping', 123, 'abc')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_mcast_rpc(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.mcast_rpc('ping', 123, 'abc')
        fc.mRPC.assert_called_once_with(1, 5, 'ping', 123, 'abc')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_mcast_rpc_opts(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.mcast_rpc('ping', 123, 'abc', group=16, ttl=3)
        fc.mRPC.assert_called_once_with(16, 3, 'ping', 123, 'abc')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_dmcast_rpc(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.dmcast_rpc(['123456', 'abcdef'], 'ping', 123, 'abc')
        fc.dmRPC.assert_called_once_with(['123456', 'abcdef'], 1, 5, 50, 'ping', 123, 'abc')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_dmcast_rpc_opts(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.dmcast_rpc(['123456', 'abcdef'], 'ping', 123, 'abc', group=16, ttl=3, delay=100)
        fc.dmRPC.assert_called_once_with(['123456', 'abcdef'], 16, 3, 100, 'ping', 123, 'abc')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_ucast_data(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.ucast_data('123456', 'Hello')
        fc.uData.assert_called_once_with('123456', 'Hello')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_ucast_data_base64(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.ucast_data('123456', 'SGVsbG8=', base64=True)
        fc.uData.assert_called_once_with('123456', 'Hello')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_mcast_data(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.mcast_data('Hello')
        fc.mData.assert_called_once_with(1, 5, 'Hello')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_mcast_data_opts(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.mcast_data('Hello', group=16, ttl=3)
        fc.mData.assert_called_once_with(16, 3, 'Hello')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_mcast_data_base64(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()
        gus.mcast_data('SGVsbG8=', base64=True)
        fc.mData.assert_called_once_with(1, 5, 'Hello')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_callback_ucast_rpc(self, base_client, future_client):
        fc = future_client.return_value
        fc.callback_rpc.return_value = helpers.make_resolved_future((True,))
        gus = self.get_gus()
        yield gus.callback_ucast_rpc('123456', 'ping', 123, 'abc')
        fc.callback_rpc.assert_called_once_with('123456', 'ping', (123, 'abc'), callback_name=None, result_type=None, retries=3, timeout=2)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_callback_ucast_rpc_opts(self, base_client, future_client):
        fc = future_client.return_value
        fc.callback_rpc.return_value = helpers.make_resolved_future((True,))
        gus = self.get_gus()
        yield gus.callback_ucast_rpc('123456', 'ping', 123, 'abc', callback_name='pong')
        fc.callback_rpc.assert_called_once_with('123456', 'ping', (123, 'abc'), callback_name='pong', result_type=None, retries=3, timeout=2)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_callback_ucast_rpc_raises_on_no_response(self, base_client, future_client):
        fc = future_client.return_value
        fc.callback_rpc.return_value = helpers.make_resolved_future(None)
        gus = self.get_gus()
        with self.assertRaises(NoResponseError):
            yield gus.callback_ucast_rpc('123456', 'ping', 123, 'abc')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_callback_mcast_rpc(self, base_client, future_client):
        fc = future_client.return_value
        fc.callback_mcast_rpc.return_value = helpers.make_resolved_future((True,))
        gus = self.get_gus()
        yield gus.callback_mcast_rpc('ping', 123, 'abc')
        fc.callback_mcast_rpc.assert_called_once_with('ping', 123, ('abc',), callback_name=None, group=1, result_type=None, retries=3, timeout=2, ttl=5)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_callback_mcast_rpc_opts(self, base_client, future_client):
        fc = future_client.return_value
        fc.callback_mcast_rpc.return_value = helpers.make_resolved_future((True,))
        gus = self.get_gus()
        yield gus.callback_mcast_rpc('ping', 123, 'abc', group=16, ttl=3, callback_name='pong')
        fc.callback_mcast_rpc.assert_called_once_with('ping', 123, ('abc',), callback_name='pong', group=16, result_type=None, retries=3, timeout=2, ttl=3)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_callback_mcast_rpc_raises_on_no_response(self, base_client, future_client):
        fc = future_client.return_value
        fc.callback_mcast_rpc.return_value = helpers.make_resolved_future(None)
        gus = self.get_gus()
        with self.assertRaises(NoResponseError):
            yield gus.callback_mcast_rpc('ping', 123, 'abc')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_callback_dmcast_rpc(self, base_client, future_client):
        fc = future_client.return_value
        fc.callback_dmcast.return_value = helpers.make_resolved_future((True,))
        gus = self.get_gus()
        yield gus.callback_dmcast_rpc(['123456', 'abcdef'], 'ping', 123, 'abc')
        fc.callback_dmcast.assert_called_once_with(['123456', 'abcdef'], 'ping', (123, 'abc'),
                                                   timeout=2, retries=3, result_type=None, callback_name=None,
                                                   dmcast_delay=50, node_limit=None, ttl=5,
                                                   chunk_results=False, previous_chunk=None, group=1)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_callback_dmcast_rpc_opts(self, base_client, future_client):
        fc = future_client.return_value
        fc.callback_dmcast.return_value = helpers.make_resolved_future((True,))
        gus = self.get_gus()
        yield gus.callback_dmcast_rpc(['123456', 'abcdef'], 'ping', 123, 'abc', group=16, ttl=3, dmcast_delay=100, callback_name='pong')
        fc.callback_dmcast.assert_called_once_with(['123456', 'abcdef'], 'ping', (123, 'abc'),
                                                   timeout=2, retries=3, result_type=None, callback_name='pong',
                                                   dmcast_delay=100, node_limit=None, ttl=3,
                                                   chunk_results=False, previous_chunk=None, group=16)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_callback_dmcast_rpc_raises_on_no_response(self, base_client, future_client):
        fc = future_client.return_value
        fc.callback_dmcast.return_value = helpers.make_resolved_future(None)
        gus = self.get_gus()
        with self.assertRaises(NoResponseError):
            yield gus.callback_dmcast_rpc(['123456', 'abcdef'], 'ping', 123, 'abc')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_traceroute(self, base_client, future_client):
        fc = future_client.return_value
        gus = self.get_gus()

        gus.traceroute('123456')
        fc.traceroute.assert_called_once_with('123456')

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_rpc_callback(self, sc, scf):
        gus = self.get_gus()
        response = Future()

        def on_ping(src, *args):
            response.set_result(True)

        gus.on_rpc('ping', on_ping)
        pkt = {'func': 'ping', 'src': '000000', 'args': []}
        gus._on_rpc_recv(pkt)
        yield with_timeout(timedelta(seconds=10), response)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_any_rpc_callback(self, sc, scf):
        gus = self.get_gus()
        response = Future()

        def on(src, func, *args):
            response.set_result(True)

        gus.on_any_rpc(on)
        pkt = {'func': 'ping', 'src': '000000', 'args': []}
        gus._on_rpc_recv(pkt)
        yield with_timeout(timedelta(seconds=10), response)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_data_callback(self, sc, scf):
        gus = self.get_gus()
        response = Future()

        def on_data(src, data):
            response.set_result(True)

        gus.on_data(on_data)
        pkt = {'src': '000000', 'data': ''}
        gus._on_data_recv(pkt)
        yield with_timeout(timedelta(seconds=10), response)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    @gen_test
    def test_traceroute_callback(self, sc, scf):
        gus = self.get_gus()
        response = Future()

        def on_traceroute(src, rtt, hops):
            response.set_result(True)

        gus.on_traceroute(on_traceroute)
        pkt = {'src': '000000', 'rtt': 23, 'hops': []}
        gus._on_traceroute_recv(pkt)
        yield with_timeout(timedelta(seconds=10), response)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_run_sync(self, sc, scf):
        gus = self.get_gus()

        @coroutine
        def do_work(gus, *args, **kwargs):
            return True

        data = gus.run_sync(10, do_work)

        self.assertTrue(data)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_has_bridge_address(self, sc, scf):
        gus = self.get_gus()
        self.assertIsNotNone(gus.bridge_address)

    @patch('snapstraction.gus.FutureClient')
    @patch('snapstraction.gus.BaseClient')
    def test_has_address(self, sc, scf):
        gus = self.get_gus()
        self.assertIsNotNone(gus.address)


class TestSCFGus(AsyncTestCase):
    """Tests for Gus SNAPconnect Futures Wrapper"""

    def get_gus(self):
        profile = SerialProfile()
        profile.address = '000020'

        network = Network()
        license = 'License.dat'

        return build_gus(profile, network, license)

    def test_init_good_profile(self):
        gus = self.get_gus()
        self.assertIsNotNone(gus)

    def test_default_nvparams_initialized(self):
        gus = self.get_gus()

        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_GROUP_INTEREST_MASK_ID, 0xFFFF)
        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_GROUP_FORWARDING_MASK_ID, 0xFFFF)
        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_GROUP_SERIAL_FORWARDING_MASK_ID, 0xFFFF)

        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID, 0)

        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_MESH_RREQ_WAIT_TIME_ID, 100)
        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_MESH_INITIAL_HOPLIMIT_ID, 3)

        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_MESH_MAX_HOPLIMIT_ID, 5)

        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_FEATURE_BITS_ID, 0x0500)

    def test_small_ttl_initialization(self):
        profile = SerialProfile()
        profile.address = '000020'

        network = Network()
        network.default_ttl = 2
        license = 'License.dat'

        gus = build_gus(profile, network, license)

        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_MESH_INITIAL_HOPLIMIT_ID, 2)
        gus._sc.save_nv_param.assert_any_call(SnapConst.NV_MESH_MAX_HOPLIMIT_ID, 2)

    @gen_test
    def test_ucast_rpc(self):
        gus = self.get_gus()
        yield gus.ucast_rpc('123456', 'ping', 123, 'abc')
        gus._scf.rpc.assert_called_once_with('123456', 'ping', 123, 'abc')

    @gen_test
    def test_mcast_rpc(self):
        gus = self.get_gus()
        yield gus.mcast_rpc('ping', 123, 'abc')
        gus._scf.mcast_rpc.assert_called_once_with('ping', 123, 'abc', group=1, ttl=5)

    @gen_test
    def test_mcast_rpc_opts(self):
        gus = self.get_gus()
        yield gus.mcast_rpc('ping', 123, 'abc', group=16, ttl=3)
        gus._scf.mcast_rpc.assert_called_once_with('ping', 123, 'abc', group=16, ttl=3)

    @gen_test
    def test_dmcast_rpc(self):
        gus = self.get_gus()
        yield gus.dmcast_rpc(['123456', '234567'], 'ping', 123, 'abc')
        gus._scf.dmcast_rpc.assert_called_once_with(['123456', '234567'], 'ping', 123, 'abc', group=1, ttl=5, delay=50)

    @gen_test
    def test_dmcast_rpc_opts(self):
        gus = self.get_gus()
        yield gus.dmcast_rpc(['123456', '234567'], 'ping', 123, 'abc', group=16, ttl=3, delay_factor=100)
        gus._scf.dmcast_rpc.assert_called_once_with(['123456', '234567'], 'ping', 123, 'abc', group=16, ttl=3, delay=100)

    @gen_test
    def test_ucast_data(self):
        gus = self.get_gus()
        yield gus.ucast_data('123456', 'Hello')
        gus._scf._snap.data_mode.assert_called_once_with('\x12\x34\x56', 'Hello')

    @gen_test
    def test_ucast_data_base64(self):
        gus = self.get_gus()
        yield gus.ucast_data('123456', 'SGVsbG8=', base64=True)
        gus._scf._snap.data_mode.assert_called_once_with('\x12\x34\x56', 'Hello')

    @gen_test
    def test_mcast_data(self):
        gus = self.get_gus()
        yield gus.mcast_data('Hello')
        gus._scf._snap.mcast_data_mode.assert_called_once_with(1, 5, 'Hello')

    @gen_test
    def test_mcast_data_opts(self):
        gus = self.get_gus()
        yield gus.mcast_data('Hello', group=16, ttl=3)
        gus._scf._snap.mcast_data_mode.assert_called_once_with(16, 3, 'Hello')

    @gen_test
    def test_mcast_data_base64(self):
        gus = self.get_gus()
        yield gus.mcast_data('SGVsbG8=', base64=True)
        gus._scf._snap.mcast_data_mode.assert_called_once_with(1, 5, 'Hello')

    @gen_test
    def test_callback_ucast_rpc(self):
        gus = self.get_gus()
        gus._scf.callback_rpc.return_value = helpers.make_resolved_future((True,))
        yield gus.callback_ucast_rpc('123456', 'ping', 123, 'abc')
        gus._scf.callback_rpc.assert_called_once_with('123456', 'ping', 123, 'abc',
                                                      validate_func=None,
                                                      callback_name=None, result_type=None, retries=0, timeout=2)

    @gen_test
    def test_callback_ucast_rpc_opts(self):
        gus = self.get_gus()
        gus._scf.callback_rpc.return_value = helpers.make_resolved_future((True,))
        yield gus.callback_ucast_rpc('123456', 'ping', 123, 'abc', callback_name='pong')
        gus._scf.callback_rpc.assert_called_once_with('123456', 'ping', 123, 'abc',
                                                      validate_func=None,
                                                      callback_name='pong', result_type=None, retries=0, timeout=2)

    @gen_test
    def test_callback_ucast_rpc_raises_on_no_response(self):
        gus = self.get_gus()
        gus._scf.callback_rpc.return_value = helpers.make_resolved_future(None)
        with self.assertRaises(NoResponseError):
            yield gus.callback_ucast_rpc('123456', 'ping', 123, 'abc')

    @gen_test
    def test_callback_mcast_rpc(self):
        gus = self.get_gus()
        gus._scf.callback_mcast_rpc.return_value = helpers.make_resolved_future((True,))
        yield gus.callback_mcast_rpc('ping', 123, 'abc')
        gus._scf.callback_mcast_rpc.assert_called_once_with('ping', 123, 'abc', callback_name=None, result_type=None,
                                                            validate_func=None,
                                                            retries=3, timeout=2, ttl=5, group=1)

    @gen_test
    def test_callback_mcast_rpc_opts(self):
        gus = self.get_gus()
        gus._scf.callback_mcast_rpc.return_value = helpers.make_resolved_future((True,))
        yield gus.callback_mcast_rpc('ping', 123, 'abc', ttl=3, group=2, callback_name='pong')
        gus._scf.callback_mcast_rpc.assert_called_once_with('ping', 123, 'abc', callback_name='pong',
                                                            validate_func=None,
                                                            result_type=None, retries=3, timeout=2, ttl=3, group=2)

    @gen_test
    def test_callback_mcast_rpc_raises_on_no_response(self):
        gus = self.get_gus()
        gus._scf.callback_mcast_rpc.return_value = helpers.make_resolved_future(None)
        with self.assertRaises(NoResponseError):
            yield gus.callback_mcast_rpc('ping', 123, 'abc')

    @gen_test
    def test_callback_dmcast_rpc(self):
        gus = self.get_gus()
        gus._scf.callback_dmcast.return_value = helpers.make_resolved_future((True,))
        yield gus.callback_dmcast_rpc(['123456', '234567'], 'ping', 123, 'abc')
        gus._scf.callback_dmcast_rpc.assert_called_once_with(['123456', '234567'], 'ping', 123, 'abc',
                                                             callback_name=None, chunk_results=True,
                                                             dmcast_delay=50, expect_seq=False,
                                                             node_limit=None, result_type=None,
                                                             validate_func=None,
                                                             retries=3, timeout=2, ttl=5, group=1)

    @gen_test
    def test_callback_dmcast_rpc_opts(self):
        gus = self.get_gus()
        gus._scf.callback_dmcast.return_value = helpers.make_resolved_future((True,))
        yield gus.callback_dmcast_rpc(['123456', '234567'], 'ping', 123, 'abc', ttl=3, group=2, callback_name='pong')
        gus._scf.callback_dmcast_rpc.assert_called_once_with(['123456', '234567'], 'ping', 123, 'abc',
                                                             callback_name='pong', chunk_results=True,
                                                             dmcast_delay=50, expect_seq=False,
                                                             node_limit=None, result_type=None,
                                                             validate_func=None,
                                                             retries=3, timeout=2, ttl=3, group=2)

    @gen_test
    def test_callback_dmcast_rpc_raises_on_no_response(self):
        gus = self.get_gus()
        gus._scf.callback_dmcast_rpc.return_value = helpers.make_resolved_future(None)
        with self.assertRaises(NoResponseError):
            yield gus.callback_dmcast_rpc(['123456', '234567'], 'ping', 123, 'abc')

    @gen_test
    def test_traceroute(self):
        gus = self.get_gus()
        gus.traceroute('123456')
        gus._scf._snap.traceroute.assert_called_once_with('\x12\x34\x56')

    @gen_test
    def test_on_rpc(self):
        gus = self.get_gus()
        response = Future()

        def on_ping(src, *args):
            response.set_result(True)

        gus.on_rpc('ping', on_ping)

        descriptor = MagicMock()
        descriptor.funcName = 'ping'
        descriptor.srcAddr = '\x12\x23\x34'
        descriptor.args = []

        gus._on_rpc_recv(descriptor)
        yield with_timeout(timedelta(seconds=10), response)

    @gen_test
    def test_on_any_rpc(self):
        gus = self.get_gus()
        response = Future()

        def on_ping(src, func, *args):
            response.set_result(True)

        gus.on_any_rpc(on_ping)

        descriptor = MagicMock()
        descriptor.funcName = 'ping'
        descriptor.srcAddr = '\x12\x23\x34'
        descriptor.args = []

        gus._on_rpc_recv(descriptor)
        yield with_timeout(timedelta(seconds=10), response)

    @gen_test
    def test_on_data(self):
        gus = self.get_gus()
        response = Future()

        def on(src, data):
            response.set_result(True)

        gus.on_data(on)

        descriptor = MagicMock()
        descriptor.funcName = 'ping'
        descriptor.srcAddr = '\x12\x23\x34'
        descriptor.args = []
        descriptor.data = "foo"

        gus._on_data_recv(descriptor)
        yield with_timeout(timedelta(seconds=10), response)

    @gen_test
    def test_on_traceroute(self):
        gus = self.get_gus()
        response = Future()

        def on(src, rtt, hops):
            response.set_result(True)

        gus.on_traceroute(on)

        gus._on_traceroute_recv('\x12\x23\x34', 34, [('\x00\x11\x22', 49)])
        yield with_timeout(timedelta(seconds=10), response)

    def test_any_rpc_multiple_calls_is_single_add(self):
        gus = self.get_gus()
        response = Future()

        def on_ping(src, func, *args):
            response.set_result(True)

        gus.on_any_rpc(on_ping)
        gus.on_any_rpc(on_ping)
        gus.on_any_rpc(on_ping)

        self.assertEqual(gus._rpc_callbacks['*'], [on_ping])

    def test_on_rpc_multiple_calls_is_single_add(self):
        gus = self.get_gus()
        response = Future()

        def on_ping(src, func, *args):
            response.set_result(True)

        gus.on_rpc('ping', on_ping)
        gus.on_rpc('ping', on_ping)
        gus.on_rpc('ping', on_ping)

        self.assertEqual(gus._rpc_callbacks['ping'], [on_ping])

    def test_on_traceroute_multiple_calls_is_single_add(self):
        gus = self.get_gus()
        response = Future()

        def on(src, rtt, hops):
            response.set_result(True)

        gus.on_traceroute(on)
        gus.on_traceroute(on)
        gus.on_traceroute(on)

        self.assertEqual(gus._traceroute_callbacks, [on])

    def test_on_data_multiple_calls_is_single_add(self):
        gus = self.get_gus()
        response = Future()

        def on(src, data):
            response.set_result(True)

        gus.on_data(on)
        gus.on_data(on)
        gus.on_data(on)

        self.assertEqual(gus._data_callbacks, [on])

    @gen_test
    def test_tx_priority_is_respected(self):
        gus = self.get_gus()
        # Default priority is middlin': 0x80
        gus.ucast_rpc('123456', 'pong')
        # Lower priority is higher
        gus.ucast_rpc('123456', 'ping', priority=0x20)

        # Even higher
        gus.ucast_rpc('123456', 'PING', priority=0x10)

        # Let things run for a bit
        yield sleep(1)

        # Check that the higher-priority call went out first
        gus._scf.rpc.assert_has_calls([
            call('123456', 'PING'),
            call('123456', 'ping'),
            call('123456', 'pong'),
        ])

    def test_tx_priority_calls_coerce(self):
        gus = self.get_gus()

        def tx_func():
            pass

        with patch("snapstraction.gus.coerce_priority") as mock_coerce:
            gus._enqueue_tx(9000, tx_func)
            mock_coerce.assert_called_once_with(9000)

    def test_coerced_tx_priority_is_enqueued(self):
        gus = self.get_gus()

        def tx_func():
            pass

        with patch.object(gus._tx_queue, "put_nowait") as mock_put:
            expected_count = gus._tx_count.next() + 1
            gus._enqueue_tx(9000, tx_func)
            mock_put.assert_called_with((255, expected_count, tx_func, ANY, False, None))

            expected_count += 1
            gus._enqueue_tx(-9000, tx_func)
            mock_put.assert_called_with((0, expected_count, tx_func, ANY, False, None))

            expected_count += 1
            gus._enqueue_tx(None, tx_func)
            mock_put.assert_called_with((0x80, expected_count, tx_func, ANY, False, None))

            expected_count += 1
            gus._enqueue_tx(256, tx_func)
            mock_put.assert_called_with((0xFF, expected_count, tx_func, ANY, False, None))

    @gen_test
    def test_tx_priority_applies_across_types(self):
        gus = self.get_gus()

        call_times = {}

        def wrap_with_timing(mock, name):
            def call_it(*args, **kwargs):
                call_times[name] = timer()
                log.debug("Mock last called: %r", call_times[name])
                return mock(*args, **kwargs)
            return call_it

        gus._scf.rpc = wrap_with_timing(gus._scf.rpc, "rpc")
        gus._scf.mcast_rpc = wrap_with_timing(gus._scf.mcast_rpc, "mcast")
        gus._scf.dmcast_rpc = wrap_with_timing(gus._scf.dmcast_rpc, "dmcast")

        # Default priority is middlin': 0x80
        gus.ucast_rpc('123456', 'pong')
        # Lower priority is higher
        gus.mcast_rpc('123456', 'ping', priority=0x20)

        # Even higher
        gus.dmcast_rpc(['123456'], 'PING', priority=0x10)

        # Let things run for a bit
        yield sleep(1)

        # assert called in order
        self.assertTrue(call_times["dmcast"] < call_times["mcast"] < call_times["rpc"])

    @gen_test(timeout=10)
    def test_tx_retries_get_higher_priority(self):
        gus = self.get_gus()
        # Pretend we get a response on the second try
        gus._scf.callback_rpc.side_effect = [
            helpers.make_resolved_future(None),
            helpers.make_resolved_future((True,)),
            helpers.make_resolved_future(None),
            helpers.make_resolved_future((True,))
        ]

        # Two requests, same initial priority...
        first = gus.callback_ucast_rpc('123456', 'ping', retries=1)
        second = gus.callback_ucast_rpc('123456', 'pong', retries=1)

        # Let things run until both requests finish
        yield second
        yield first

        # Check that the higher-priority call went out first, along with its retry.
        gus._scf.callback_rpc.assert_has_calls([
            call('123456', 'ping', callback_name=None, result_type=None, retries=0, timeout=2, validate_func=None),
            call('123456', 'ping', callback_name=None, result_type=None, retries=0, timeout=2, validate_func=None),
            call('123456', 'pong', callback_name=None, result_type=None, retries=0, timeout=2, validate_func=None),
            call('123456', 'pong', callback_name=None, result_type=None, retries=0, timeout=2, validate_func=None)])

    @gen_test
    def test_high_priority_multi_command_behavior(self):
        gus = self.get_gus()

        # Using non-callback forms to simplify the test, normally these
        # operations would use callback_ucast_rpc...

        @coroutine
        def move_node(target):
            # High Priority!
            yield gus.ucast_rpc(target, 'saveNvParam', 4, 1, priority=0x10)
            yield gus.ucast_rpc(target, 'saveNvParam', 3, 0xD110, priority=0x10)

        @coroutine
        def another_op(target):
            # Default Priority
            yield gus.ucast_rpc(target, 'ping')
            yield gus.ucast_rpc(target, 'pong')

        # Pretend "another_op" got queued up elsewhere
        op = another_op('123456')

        # And a high priority operation needs to run
        for t in ['123456', '7890ab', 'cdefed']:
            yield move_node(t)

        yield op

        # What should happen is the high priority task fully completes,
        # in order, before the lower priority task
        gus._scf.rpc.assert_has_calls([
            call('123456', 'saveNvParam', 4, 1),
            call('123456', 'saveNvParam', 3, 0xd110),
            call('7890ab', 'saveNvParam', 4, 1),
            call('7890ab', 'saveNvParam', 3, 0xd110),
            call('cdefed', 'saveNvParam', 4, 1),
            call('cdefed', 'saveNvParam', 3, 0xd110),
            call('123456', 'ping'),
            call('123456', 'pong'),
        ])

    @gen_test
    def test_metrics_exist(self):
        gus = self.get_gus()

        @coroutine
        def do_a_thing(target):
            # Default Priority
            yield gus.ucast_rpc(target, 'ping')
            yield gus.ucast_rpc(target, 'pong')

        op = do_a_thing('123456')
        yield op

        self.assertEqual(0, gus._metrics[snapstraction.gus.METRIC_PACKET_RX])
        self.assertEqual(2, gus._metrics[snapstraction.gus.METRIC_PACKET_TX])
        self.assertEqual(0, gus._metrics[snapstraction.gus.METRIC_TX_EXCEPTION])
        self.assertGreater(gus._metrics[snapstraction.gus.METRIC_UPTIME], 0)
        self.assertGreater(gus._metrics[snapstraction.gus.METRIC_TX_TOTAL_TIME], 0)
        self.assertGreaterEqual(gus._metrics[snapstraction.gus.METRIC_TX_MEAN_TIME], 0)
        self.assertGreater(gus._metrics[snapstraction.gus.METRIC_TX_TOTAL_WAIT_TIME], 0)
        self.assertGreater(gus._metrics[snapstraction.gus.METRIC_TX_MEAN_WAIT_TIME], 0)
        self.assertEqual(0, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])

    @gen_test
    def test_ewma_metric_update(self):
        gus = self.get_gus()

        # Uninitialized is None
        self.assertEqual(None, gus._metrics[snapstraction.gus.METRIC_TX_MEAN_TIME])
        gus._update_ewma_metric(snapstraction.gus.METRIC_TX_MEAN_TIME, 1)
        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_MEAN_TIME])
        gus._update_ewma_metric(snapstraction.gus.METRIC_TX_MEAN_TIME, 10)
        self.assertEqual(5, gus._metrics[snapstraction.gus.METRIC_TX_MEAN_TIME])

    @gen_test
    def test_rx_increments_metric(self):
        gus = self.get_gus()
        self.assertEqual(0, gus._metrics[snapstraction.gus.METRIC_PACKET_RX])

        descriptor = MagicMock()
        descriptor.funcName = 'ping'
        descriptor.srcAddr = '\x12\x23\x34'
        descriptor.args = []
        gus._on_rpc_recv(descriptor)
        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_PACKET_RX])

        descriptor = MagicMock()
        descriptor.srcAddr = '\x12\x23\x34'
        descriptor.data = "foo"
        gus._on_data_recv(descriptor)
        self.assertEqual(2, gus._metrics[snapstraction.gus.METRIC_PACKET_RX])

        gus._on_traceroute_recv('\x12\x23\x34', 34, [('\x00\x11\x22', 49)])
        self.assertEqual(3, gus._metrics[snapstraction.gus.METRIC_PACKET_RX])

    @gen_test
    def test_tx_exception_increments_metric(self):
        gus = self.get_gus()

        self.assertEqual(0, gus._metrics[snapstraction.gus.METRIC_TX_EXCEPTION])

        @coroutine
        def bad_tx():
            raise Exception("This is fine")

        gus._enqueue_tx(0x80, bad_tx)

        yield sleep(.1)

        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_EXCEPTION])

    @gen_test
    def test_gus_has_arbiter_hook(self):
        gus = self.get_gus()
        self.assertIsNone(gus.arbiter)
        self.assertEqual(gus.arbiter_timeout, timedelta(seconds=2))

    @gen_test
    def test_gus_tx_calls_arbiter(self):
        gus = self.get_gus()

        gus.arbiter = Arbiter()
        gus.arbiter.acquire.returns(True)

        self.tx_called = False

        @coroutine
        def tx():
            self.tx_called = True

        gus._enqueue_tx(0x80, tx)

        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])

        yield sleep(.1)

        gus.arbiter.acquire.assert_called_correctly()
        self.assertTrue(self.tx_called)
        self.assertEqual(0, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])
        gus.arbiter.release.assert_called_correctly()

    @gen_test
    def test_unicast_to_bridge_bypasses_arbiter(self):
        gus = self.get_gus()

        gus.arbiter = Arbiter()
        gus.arbiter.acquire.returns(True)

        self.tx_called = False

        @coroutine
        def tx():
            self.tx_called = True

        gus._enqueue_tx(0x80, tx, True, gus.bridge_addr)

        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])

        yield sleep(.1)

        gus.arbiter.acquire.assert_not_called()
        self.assertTrue(self.tx_called)
        self.assertEqual(0, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])
        gus.arbiter.release.assert_not_called()

    @gen_test
    def test_gus_tx_accepts_failure(self):
        gus = self.get_gus()

        gus.arbiter = Arbiter()
        gus.arbiter.acquire.returns(False)

        self.tx_called = False

        @coroutine
        def tx():
            self.tx_called = True

        gus._enqueue_tx(0x80, tx)

        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])

        yield sleep(.1)

        gus.arbiter.acquire.assert_called_correctly()
        self.assertFalse(self.tx_called)
        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])
        gus.arbiter.release.assert_not_called()

    @gen_test
    def test_gus_tx_accepts_exceptional_failure(self):
        gus = self.get_gus()

        gus.arbiter = Arbiter()
        gus.arbiter.acquire.raises(Exception)

        self.tx_called = False

        @coroutine
        def tx():
            self.tx_called = True

        gus._enqueue_tx(0x80, tx)

        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])

        yield sleep(.1)

        gus.arbiter.acquire.assert_called_correctly()
        self.assertFalse(self.tx_called)
        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])
        gus.arbiter.release.assert_not_called()

    @gen_test
    def test_gus_tx_handles_timeout(self):
        gus = self.get_gus()

        gus.arbiter = Arbiter()
        gus.arbiter_timeout = timedelta(seconds=.1)
        gus.arbiter.acquire.returns(True, .2)

        self.tx_called = False

        @coroutine
        def tx():
            self.tx_called = True

        gus._enqueue_tx(0x80, tx)
        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])

        yield sleep(.5)

        gus._running = False

        yield sleep(.1)

        gus.arbiter.acquire.assert_called_correctly()
        self.assertFalse(self.tx_called)
        self.assertEqual(1, gus._metrics[snapstraction.gus.METRIC_TX_QUEUE_LENGTH])
        gus.arbiter.release.assert_not_called()


class TestSerialGus(AsyncTestCase):
    """Tests for Gus Serial Wrapper

    Most functionality of the SCFGus is tested in the TestSCFGus class...
    """

    def get_gus(self):
        profile = SerialProfile()
        profile.address = '000020'

        network = Network()
        license = 'License.dat'

        gus = Gus(profile, network, license)
        gus._schedule_tx_service()

        sc = MagicMock()
        sc.licensed_mac = '\x12\x34\x56'

        scf = MagicMock()
        scf.rpc.return_value = helpers.make_resolved_future((None,))
        scf.mcast_rpc.return_value = helpers.make_resolved_future((None,))
        scf.dmcast_rpc.return_value = helpers.make_resolved_future((None,))

        scf.callback_rpc.return_value = helpers.make_resolved_future((None,))
        mcast_result = MultiResult({}, {})
        mcast_result.retry = MagicMock()
        mcast_result.retry.return_value = helpers.make_resolved_future(None)
        scf.callback_mcast_rpc.return_value = helpers.make_resolved_future(mcast_result)
        dmcast_result = MultiResult({'123456': (True,)}, {})
        dmcast_result.retry = MagicMock()
        dmcast_result.retry.return_value = helpers.make_resolved_future(None)
        scf.callback_dmcast_rpc.return_value = helpers.make_resolved_future(dmcast_result)

        scf._snap = sc

        gus._scf = scf
        gus._sc = sc

        return gus

    def test_init_good_profile(self):
        gus = self.get_gus()
        self.assertIsNotNone(gus)

    def test_has_bridge_address(self):
        gus = self.get_gus()
        try:
            gus.bridge_address
        except:
            log.exception("Unexpected exception")
            self.fail("Unexpected exception")

    def test_has_address(self):
        gus = self.get_gus()
        self.assertIsNotNone(gus.address)

    def test_start_calls_open_serial(self):
        gus = self.get_gus()
        gus._scf.open_serial.return_value = helpers.make_resolved_future((gus.bridge_address,))
        gus.start()
        gus._scf.open_serial.assert_called_once_with(1, "/dev/snap1", baudrate=38400)

    def test_start_polls_asyncore(self):
        gus = self.get_gus()
        gus._scf.open_serial.return_value = helpers.make_resolved_future((gus.bridge_address,))
        gus._asyncpoll = MagicMock()
        gus.start()
        self.assertEqual(gus._asyncpoll.mock_calls, [mock.call.start()])

    def test_stop_calls_close_serial(self):
        gus = self.get_gus()
        gus._scf.close_serial.return_value = helpers.make_resolved_future(None)
        gus.stop()
        gus._scf.close_serial.assert_called_once_with(1, "/dev/snap1")

    def test_stop_stops_asyncore_poll(self):
        gus = self.get_gus()
        gus._scf.close_serial.return_value = helpers.make_resolved_future(None)
        gus._asyncpoll = MagicMock()
        gus.stop()
        self.assertEqual(gus._asyncpoll.mock_calls, [mock.call.stop()])

class TestTCPGus(AsyncTestCase):
    """Tests for Gus TCP Wrapper

    Most functionality of the SCFGus is tested in the TestSCFGus class...
    """

    def get_gus(self):
        profile = TcpProfile()
        profile.address = '000020'

        network = Network()
        license = 'License.dat'

        gus = Gus(profile, network, license)
        gus._schedule_tx_service()

        sc = MagicMock()
        sc.licensed_mac = '\x12\x34\x56'

        scf = MagicMock()
        scf.rpc.return_value = helpers.make_resolved_future((None,))
        scf.mcast_rpc.return_value = helpers.make_resolved_future((None,))
        scf.dmcast_rpc.return_value = helpers.make_resolved_future((None,))

        scf.callback_rpc.return_value = helpers.make_resolved_future((None,))
        mcast_result = MultiResult({}, {})
        mcast_result.retry = MagicMock()
        mcast_result.retry.return_value = helpers.make_resolved_future(None)
        scf.callback_mcast_rpc.return_value = helpers.make_resolved_future(mcast_result)
        dmcast_result = MultiResult({'123456': (True,)}, {})
        dmcast_result.retry = MagicMock()
        dmcast_result.retry.return_value = helpers.make_resolved_future(None)
        scf.callback_dmcast_rpc.return_value = helpers.make_resolved_future(dmcast_result)

        scf._snap = sc

        gus._scf = scf
        gus._sc = sc

        return gus

    def test_init_good_profile(self):
        gus = self.get_gus()
        self.assertIsNotNone(gus)

    def test_has_bridge_address(self):
        gus = self.get_gus()
        try:
            gus.bridge_address
        except:
            log.exception("Unexpected exception")
            self.fail("Unexpected exception")

    def test_has_address(self):
        gus = self.get_gus()
        self.assertIsNotNone(gus.address)


class TestPriorityCoercion(unittest.TestCase):
    def test_priority_high(self):
        self.assertEqual(0, coerce_priority(0))
        self.assertEqual(0, coerce_priority(-15))
        self.assertEqual(1, coerce_priority(1))

    def test_priority_medium(self):
        self.assertEqual(10, coerce_priority(10))
        self.assertEqual(64, coerce_priority(0x40))
        self.assertEqual(128, coerce_priority(0x80))

    def test_priority_low(self):
        self.assertEqual(255, coerce_priority(0xFF))
        self.assertEqual(255, coerce_priority(0x100))
        self.assertEqual(200, coerce_priority(200))

    def test_priority_bad(self):
        self.assertEqual(0x80, coerce_priority(None))
        self.assertEqual(0x80, coerce_priority("low"))
        self.assertEqual(0x80, coerce_priority("high"))
        self.assertEqual(99, coerce_priority(99.256))
        self.assertEqual(0x80, coerce_priority(True))
        self.assertEqual(0x80, coerce_priority(False))
