import binascii
import logging
import mock

from collections import OrderedDict

from tornado.testing import AsyncTestCase, gen_test
from tornado.concurrent import Future
from tornado.gen import sleep

from tests import helpers
from preggy import expect

from snapstraction import mesh
from snapstraction.gus import Gus

LOG = logging.getLogger(__name__)


class MockGus(Gus):
    def __new__(cls):
        obj = object.__new__(Gus)
        obj.__init__()
        return obj


def mock_gus():
    gus = mock.MagicMock()
    gus.bridge_address = '000001'
    gus.address = '111110'

    gus.ucast_rpc.return_value = Future()
    gus.ucast_rpc.return_value.set_result(True)

    gus.mcast_rpc.return_value = Future()
    gus.mcast_rpc.return_value.set_result(True)

    gus.dmcast_rpc.return_value = Future()
    gus.dmcast_rpc.return_value.set_result(True)

    gus.callback_ucast_rpc.return_value = Future()
    gus.callback_ucast_rpc.return_value.set_result(True)

    gus.callback_mcast_rpc.return_value = Future()
    gus.callback_mcast_rpc.return_value.set_result(True)

    gus.callback_dmcast_rpc.return_value = Future()
    gus.callback_dmcast_rpc.return_value.set_result(True)

    gus.ucast_data.return_value = Future()
    gus.ucast_data.return_value.set_result(True)

    gus.mcast_data.return_value = Future()
    gus.mcast_data.return_value.set_result(True)

    gus.callback_ucast_data.return_value = Future()
    gus.callback_ucast_data.return_value.set_result(True)

    gus.callback_mcast_data.return_value = Future()
    gus.callback_mcast_data.return_value.set_result(True)

    gus.traceroute.return_value = Future()
    gus.traceroute.return_value.set_result(True)

    gus._base = MockGus()

    gus.on_rpc = gus._base.on_rpc
    gus.on_any_rpc = gus._base.on_any_rpc
    gus.on_data = gus._base.on_data
    gus.on_traceroute = gus._base.on_traceroute

    return gus


class MeshTestCase(AsyncTestCase):

    def setUp(self):
        super(MeshTestCase, self).setUp()

        self.gus = mock_gus()

        self.target = 'aabbcc'  # The address of the node to interact with
        self.snap_address = '012345'  # The address of the snap belt node
        self.pkt_content = 'foobar'  # The payload of incoming data packets

        self.gus.address = self.snap_address

    @gen_test
    def test_listen_for_intercepted_output(self):
        output_func = mock.Mock()

        mesh.listen_for_intercepted_output(self.gus, output_func, _timeout=0.1)
        self.gus._base._on_data_recv(self.target, self.pkt_content)

        yield sleep(.2)  # Let the loop run, briefly...

        output_func.assert_called_once_with({'node': self.target, 'packet': self.pkt_content})

    @gen_test
    def test_listen_for_data_messages(self):
        output_func = mock.Mock()

        mesh.listen(self.gus, output_func, _timeout=0.1)
        self.gus._base._on_data_recv(self.target, self.pkt_content)

        yield sleep(.2)  # Let the loop run, briefly...

        output_func.assert_called_once_with({'src': self.target, 'data': self.pkt_content})

    @gen_test
    def test_listen_for_rpc_messages(self):
        output_func = mock.Mock()

        mesh.listen(self.gus, output_func, _timeout=0.1)
        self.gus._base._on_rpc_recv(self.target, 'ping', 1, False, None, 'abc')

        yield sleep(.2)  # Let the loop run, briefly...

        output_func.assert_called_once_with({'src': self.target, 'func': 'ping', 'args': (1, False, None, 'abc')})

    @gen_test
    def test_unicast_ping(self):
        count = 2
        ttl = 5

        data = yield mesh.unicast_ping(self.gus, count=count, ttl=ttl)
        self.gus._base._on_rpc_recv(self.target, 'tellVmStat', 5 + (42 << 8), 'abc')

        yield sleep(.2)  # Let the loop run, briefly...

        self.assertIn(self.target, data)
        self.assertEqual(data[self.target], 42)

    @gen_test
    def test_unicast_ping_handles_repeated_responses(self):
        count = 2
        ttl = 5

        progress_cb = mock.MagicMock()

        result = mesh.unicast_ping(self.gus, count=count, ttl=ttl, progress_cb=progress_cb)

        self.gus._base._on_rpc_recv(self.target, 'tellVmStat', 5 + (42 << 8), 'abc')
        self.gus._base._on_rpc_recv(self.target, 'tellVmStat', 5 + (99 << 8), 'abc')

        data = yield result

        self.assertIn(self.target, data)
        self.assertEqual(1, len(data))
        # We only track the first lq right now
        self.assertEqual(data[self.target], 42)
        progress_cb.assert_has_calls([mock.call.__nonzero__(),
                                      mock.call(0, 1, self.target),
                                      mock.call.__nonzero__(),
                                      mock.call(1, 1, self.target)])

    @gen_test
    def test_dmcast_ping(self):
        count = 2
        ttl = 5

        data = yield mesh.dmcast_ping(self.gus, count=count, ttl=ttl, response_name="dmPing000000")
        self.gus._base._on_rpc_recv(self.target, 'dmPing000000', 42)

        yield sleep(.2)  # Let the loop run, briefly...

        self.assertIn(self.target, data)
        self.assertEqual(data[self.target], 42)

    @gen_test
    def test_dmcast_ping_handles_repeated_responses(self):
        count = 2
        ttl = 5

        progress_cb = mock.MagicMock()

        result = mesh.dmcast_ping(self.gus, count=count, ttl=ttl, response_name="dmPing000000", progress_cb=progress_cb)
        self.gus._base._on_rpc_recv(self.target, 'dmPing000000', 42)
        self.gus._base._on_rpc_recv(self.target, 'dmPing000000', 42)

        data = yield result  # Let the loop run, briefly...

        self.assertIn(self.target, data)
        self.assertEqual(1, len(data))
        # We only track the first lq right now
        self.assertEqual(data[self.target], 42)
        progress_cb.assert_has_calls([mock.call.__nonzero__(),
                                      mock.call(0, 1, self.target),
                                      mock.call.__nonzero__(),
                                      mock.call(1, 1, self.target)])
