import binascii
from itertools import chain
import mock
from os import path
import unittest
from tornado.testing import AsyncTestCase, gen_test
from tornado.concurrent import Future

import snaplib.ScriptsManager
from snappy import SnapConstantsHelper
from snapstraction import node
from snapstraction import nvparam
from snapstraction.gus import Gus, NoResponseError
from tests import helpers


class MockGus(Gus):
    def __new__(cls):
        obj = object.__new__(Gus)
        obj.__init__()
        return obj


def mock_gus():
    gus = mock.MagicMock()
    gus.bridge_address = '000001'
    gus.address = '111110'

    gus.network.default_cb_timeout = 2
    gus.network.default_cb_retries = 3

    gus.ucast_rpc.return_value = Future()
    gus.ucast_rpc.return_value.set_result(True)

    gus.mcast_rpc.return_value = Future()
    gus.mcast_rpc.return_value.set_result(True)

    gus.dmcast_rpc.return_value = Future()
    gus.dmcast_rpc.return_value.set_result(True)

    gus.callback_ucast_rpc.return_value = Future()
    gus.callback_ucast_rpc.return_value.set_result(True)

    gus.callback_mcast_rpc.return_value = Future()
    gus.callback_mcast_rpc.return_value.set_result(True)

    gus.callback_dmcast_rpc.return_value = Future()
    gus.callback_dmcast_rpc.return_value.set_result(True)

    gus.ucast_data.return_value = Future()
    gus.ucast_data.return_value.set_result(True)

    gus.mcast_data.return_value = Future()
    gus.mcast_data.return_value.set_result(True)

    gus.callback_ucast_data.return_value = Future()
    gus.callback_ucast_data.return_value.set_result(True)

    gus.callback_mcast_data.return_value = Future()
    gus.callback_mcast_data.return_value.set_result(True)

    gus.traceroute.return_value = Future()
    gus.traceroute.return_value.set_result(True)

    gus._base = MockGus()

    gus.on_rpc = gus._base.on_rpc
    gus.on_any_rpc = gus._base.on_any_rpc
    gus.on_data = gus._base.on_data
    gus.on_traceroute = gus._base.on_traceroute

    # Setting this for consistency when we do image load tests.
    # If this is True, chunk sizes will be 91 instead of 93 below.
    SnapConstantsHelper._packetCRCEnable = False

    return gus


class BaseTestCase(unittest.TestCase):

    def setUp(self):
        self.gus = mock_gus()

        self.target = 'aabbcc'  # The address of the node to interact with
        self.snap_address = '012345'  # The address of the snap belt node
        self.pkt_content = 'foobar'  # The payload of incoming data packets

        self.gus.address = self.snap_address


class NodeTestCase(AsyncTestCase, BaseTestCase):

    @gen_test
    def test_get_nvparam_by_id(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future(('Bridge', ))

        data = yield nvparam.get_nvparam(self.gus, self.target, '8')

        self.assertEqual(data, {'nvparam device.name': 'Bridge'})
        self.gus.callback_ucast_rpc.assert_called_once_with(self.target, 'loadNvParam', 8)

    @gen_test
    def test_get_nvparam_by_id_no_name(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future(('Something Useful', ))

        # NVParam 200 is in the USER area.
        data = yield nvparam.get_nvparam(self.gus, self.target, '200')

        self.assertEqual(data, {'nvparam 200': 'Something Useful'})
        self.gus.callback_ucast_rpc.assert_called_once_with(self.target, 'loadNvParam', 200)

    @gen_test
    def test_get_nvparam_by_name(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future(('Bridge',))

        data = yield nvparam.get_nvparam(self.gus, self.target, 'device.name')

        self.assertEqual(data, {'nvparam device.name': 'Bridge'})
        self.gus.callback_ucast_rpc.assert_called_once_with(self.target, 'loadNvParam', 8)

    @gen_test
    def test_check_and_set_raises_out_of_retries(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future(('None', ))

        with self.assertRaises(Exception):
            yield nvparam.check_and_set_nvparam(self.gus, self.target, 200, '8')

    def test_parse_nvparam_raises_on_bad_name(self):
        with self.assertRaises(ValueError):
            nvparam.parse_nvparam('Alas poor Yorick, I knew him well...')

    @gen_test
    def test_get_nvparam_formatted(self):
        nvparams = [
            ('net.network-id', 0x1c2c, '0x1c2c'),
            ('device.mac-address', '\x01\x23\x45\x67\x89\xab\xcd\xef', '01:23:45:67:89:ab:cd:ef'),
            ('device.feature-bits', 0x7f, '0x007f'),
            ('device.snappy-crc', 0xdead, '0xdead'),
            ('device.vendor-config-bits', 0xbeef, '0xbeef'),
            ('mesh.override', 1, True),
            ('security.enable-encryption', 2, 'Basic'),
            ('security.lockdown-bitmask', 0, '0x0000'),
        ]

        for (param_name, raw_value, formatted_value) in nvparams:
            self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future((raw_value,))

            data = yield nvparam.get_nvparam(self.gus, self.target, param_name)

            self.assertEqual(data, {'nvparam %s' % param_name: formatted_value})

    @gen_test
    def test_get_nvparam_feature_bit_on(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future((0x1FF,))

        data = yield nvparam.get_nvparam(self.gus, self.target, 'device.feature-bits.enable-second-crc')

        self.assertEqual(data, {'nvparam device.feature-bits.enable-second-crc': True})
        self.gus.callback_ucast_rpc.assert_called_once_with(self.target, 'loadNvParam', 11)

    @gen_test
    def test_get_nvparam_feature_bit_off(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future((0x0FF,))

        data = yield nvparam.get_nvparam(self.gus, self.target, 'device.feature-bits.enable-second-crc')

        self.assertEqual(data, {'nvparam device.feature-bits.enable-second-crc': False})
        self.gus.callback_ucast_rpc.assert_called_once_with(self.target, 'loadNvParam', 11)

    @gen_test
    def test_set_nvparam_by_id(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future(('Bridge',))
        ]

        yield nvparam.set_nvparam(self.gus, self.target, '8', 'Bridge')

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'saveNvParam', 8, 'Bridge'),
            mock.call(self.target, 'loadNvParam', 8)
        ])

    @gen_test
    def test_set_nvparam_by_name(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future(('Bridge',))
        ]

        yield nvparam.set_nvparam(self.gus, self.target, 'device.name', 'Bridge')

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'saveNvParam', 8, 'Bridge'),
            mock.call(self.target, 'loadNvParam', 8)
        ])

    @gen_test
    def test_set_nvparam_feature_bit_on(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((0x0FF,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((0x1FF,)),
        ]

        yield nvparam.set_nvparam(self.gus, self.target, 'device.feature-bits.enable-second-crc', True)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 11),
            mock.call(self.target, 'saveNvParam', 11, 0x1FF),
            mock.call(self.target, 'loadNvParam', 11)
        ])

    @gen_test
    def test_set_nvparam_feature_bit_off(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((0x1FF,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((0x0FF,)),
        ]

        yield nvparam.set_nvparam(self.gus, self.target, 'device.feature-bits.enable-second-crc', False)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 11),
            mock.call(self.target, 'saveNvParam', 11, 0x0FF),
            mock.call(self.target, 'loadNvParam', 11)
        ])

    @gen_test
    def test_reset_nvparams_ota(self):
        yield nvparam.reset_nvparams(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_called_once_with(self.target, 'saveNvParam', 1, 0)

    @gen_test
    def test_start_output_interceptor_all(self):
        yield node.start_output_interceptor(self.gus, self.target, True, True)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'uniConnect', SnapConstantsHelper.DS_TRANSPARENT, SnapConstantsHelper.DS_STDIO),
            mock.call(self.target, 'uniConnect', SnapConstantsHelper.DS_TRANSPARENT, SnapConstantsHelper.DS_ERROR),
            mock.call(self.target, 'ucastSerial', binascii.unhexlify(self.gus.address))
        ])

    @gen_test
    def test_start_output_interceptor_stdout(self):
        yield node.start_output_interceptor(self.gus, self.target, True, False)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'uniConnect', SnapConstantsHelper.DS_TRANSPARENT, SnapConstantsHelper.DS_STDIO),
            mock.call(self.target, 'ucastSerial', binascii.unhexlify(self.gus.address))
        ])

    @gen_test
    def test_start_output_interceptor_stderr(self):
        yield node.start_output_interceptor(self.gus, self.target, False, True)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'uniConnect', SnapConstantsHelper.DS_TRANSPARENT, SnapConstantsHelper.DS_ERROR),
            mock.call(self.target, 'ucastSerial', binascii.unhexlify(self.gus.address))
        ])

    @gen_test
    def test_stop_output_interceptor(self):
        yield node.stop_output_interceptor(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'uniConnect', SnapConstantsHelper.DS_NULL, SnapConstantsHelper.DS_STDIO),
            mock.call(self.target, 'uniConnect', SnapConstantsHelper.DS_NULL, SnapConstantsHelper.DS_ERROR)
        ])

    # Not a gen_test because this is just checking that we send the underlying traceroute.
    def test_traceroute(self):
        node.traceroute(self.gus, self.target)

        self.gus.traceroute.assert_called_once_with(self.target)

    @gen_test
    def test_reboot(self):
        yield node.reboot(self.gus, self.target)

        self.gus.callback_ucast_rpc.called_once_with(self.target, 'reboot')

    @gen_test
    def test_get_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future(('\x0d\xec\xaf',)),
            helpers.make_resolved_future(4),
            helpers.make_resolved_future(0x1c2c),
            helpers.make_resolved_future((4,)),
            helpers.make_resolved_future((0x1c2c,)),
            helpers.make_resolved_future((0x011f,)),
            helpers.make_resolved_future((0x0000,)),
        ]

        yield node.get_info(self.gus, self.target, verbosity=0)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 2),
            mock.call(self.target, 'getChannel'),
            mock.call(self.target, 'getNetId'),
            mock.call(self.target, 'loadNvParam', 4),
            mock.call(self.target, 'loadNvParam', 3),
            mock.call(self.target, 'loadNvParam', 11),
            mock.call(self.target, 'loadNvParam', 64)])

    @gen_test
    def test_get_info_verbose(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future(('\x0d\xec\xaf',)),
            helpers.make_resolved_future(4),
            helpers.make_resolved_future(0x1c2c),
            helpers.make_resolved_future((4,)),
            helpers.make_resolved_future((0x1c2c,)),
            helpers.make_resolved_future((0x011f,)),
            helpers.make_resolved_future((0x0000,)),
            helpers.make_resolved_future(('DeviceName',)),
            helpers.make_resolved_future(('DeviceType',)),
            helpers.make_resolved_future(('ScriptName',)),
            helpers.make_resolved_future((0xC4C4,)),
            helpers.make_resolved_future((8, 0xc0de)),
            helpers.make_resolved_future((5,)),
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,27,32767,SM220')),
            helpers.make_resolved_future((27,))
        ]

        yield node.get_info(self.gus, self.target, verbosity=1)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 2),
            mock.call(self.target, 'getChannel'),
            mock.call(self.target, 'getNetId'),
            mock.call(self.target, 'loadNvParam', 4),
            mock.call(self.target, 'loadNvParam', 3),
            mock.call(self.target, 'loadNvParam', 11),
            mock.call(self.target, 'loadNvParam', 64),
            mock.call(self.target, 'loadNvParam', 8),
            mock.call(self.target, 'loadNvParam', 10),
            mock.call(self.target, 'imageName'),
            mock.call(self.target, 'loadNvParam', 40),
            mock.call(self.target, 'vmStat', 8, callback_name='tellVmStat'),
            mock.call(self.target, 'loadNvParam', 41),
            mock.call(self.target, 'vmStat', 6, callback_name='tellVmStat'),
            mock.call(self.target, 'getInfo', 3)
        ])

    @gen_test
    def test_get_info_more_verbose(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future(('\x0d\xec\xaf',)),
            helpers.make_resolved_future(4),
            helpers.make_resolved_future(0x1c2c),
            helpers.make_resolved_future((4,)),
            helpers.make_resolved_future((0x1c2c,)),
            helpers.make_resolved_future((0x011f,)),
            helpers.make_resolved_future((0x0000,)),
            helpers.make_resolved_future(('DeviceName',)),
            helpers.make_resolved_future(('DeviceType',)),
            helpers.make_resolved_future(('ScriptName',)),
            helpers.make_resolved_future((0xC4C4,)),
            helpers.make_resolved_future((8, 0xc0de)),
            helpers.make_resolved_future((5,)),
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,27,32767,SM220')),
            helpers.make_resolved_future((27,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((False,)),
            helpers.make_resolved_future((False,)),
            helpers.make_resolved_future((True,)),
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((None,)),
        ]

        yield node.get_info(self.gus, self.target, verbosity=2)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 2),
            mock.call(self.target, 'getChannel'),
            mock.call(self.target, 'getNetId'),
            mock.call(self.target, 'loadNvParam', 4),
            mock.call(self.target, 'loadNvParam', 3),
            mock.call(self.target, 'loadNvParam', 11),
            mock.call(self.target, 'loadNvParam', 64),
            mock.call(self.target, 'loadNvParam', 8),
            mock.call(self.target, 'loadNvParam', 10),
            mock.call(self.target, 'imageName'),
            mock.call(self.target, 'loadNvParam', 40),
            mock.call(self.target, 'vmStat', 8, callback_name='tellVmStat'),
            mock.call(self.target, 'loadNvParam', 41),
            mock.call(self.target, 'vmStat', 6, callback_name='tellVmStat'),
            mock.call(self.target, 'getInfo', 3),
            mock.call(self.target, 'loadNvParam', 19),
            mock.call(self.target, 'loadNvParam', 39),
            mock.call(self.target, 'loadNvParam', 12),
            mock.call(self.target, 'loadNvParam', 13),
            mock.call(self.target, 'loadNvParam', 14),
            mock.call(self.target, 'loadNvParam', 15),
            mock.call(self.target, 'loadNvParam', 80),
            mock.call(self.target, 'loadNvParam', 81),
            mock.call(self.target, 'loadNvParam', 16),
            mock.call(self.target, 'loadNvParam', 17),
            mock.call(self.target, 'loadNvParam', 18),
            mock.call(self.target, 'loadNvParam', 5),
            mock.call(self.target, 'loadNvParam', 6),
            mock.call(self.target, 'loadNvParam', 33),
            mock.call(self.target, 'loadNvParam', 91),
            mock.call(self.target, 'loadNvParam', 78)])

    @gen_test
    def test_get_info_even_more_verbose(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future(('\x0d\xec\xaf',)),
            helpers.make_resolved_future(4),
            helpers.make_resolved_future(0x1c2c),
            helpers.make_resolved_future((4,)),
            helpers.make_resolved_future((0x1c2c,)),
            helpers.make_resolved_future((0x011f,)),
            helpers.make_resolved_future((0x0000,)),
            helpers.make_resolved_future(('DeviceName',)),
            helpers.make_resolved_future(('DeviceType',)),
            helpers.make_resolved_future(('ScriptName',)),
            helpers.make_resolved_future((0xC4C4,)),
            helpers.make_resolved_future((8, 0xc0de)),
            helpers.make_resolved_future((5,)),
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,27,32767,SM220')),
            helpers.make_resolved_future((27,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((False,)),
            helpers.make_resolved_future((False,)),
            helpers.make_resolved_future((True,)),
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((None,)),

            helpers.make_resolved_future((1000,)),
            helpers.make_resolved_future((1000,)),
            helpers.make_resolved_future((1000,)),
            helpers.make_resolved_future((1000,)),
            helpers.make_resolved_future((1000,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((0,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
        ]

        yield node.get_info(self.gus, self.target, verbosity=3)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 2),
            mock.call(self.target, 'getChannel'),
            mock.call(self.target, 'getNetId'),
            mock.call(self.target, 'loadNvParam', 4),
            mock.call(self.target, 'loadNvParam', 3),
            mock.call(self.target, 'loadNvParam', 11),
            mock.call(self.target, 'loadNvParam', 64),
            mock.call(self.target, 'loadNvParam', 8),
            mock.call(self.target, 'loadNvParam', 10),
            mock.call(self.target, 'imageName'),
            mock.call(self.target, 'loadNvParam', 40),
            mock.call(self.target, 'vmStat', 8, callback_name='tellVmStat'),
            mock.call(self.target, 'loadNvParam', 41),
            mock.call(self.target, 'vmStat', 6, callback_name='tellVmStat'),
            mock.call(self.target, 'getInfo', 3),
            mock.call(self.target, 'loadNvParam', 19),
            mock.call(self.target, 'loadNvParam', 39),
            mock.call(self.target, 'loadNvParam', 12),
            mock.call(self.target, 'loadNvParam', 13),
            mock.call(self.target, 'loadNvParam', 14),
            mock.call(self.target, 'loadNvParam', 15),
            mock.call(self.target, 'loadNvParam', 80),
            mock.call(self.target, 'loadNvParam', 81),
            mock.call(self.target, 'loadNvParam', 16),
            mock.call(self.target, 'loadNvParam', 17),
            mock.call(self.target, 'loadNvParam', 18),
            mock.call(self.target, 'loadNvParam', 5),
            mock.call(self.target, 'loadNvParam', 6),
            mock.call(self.target, 'loadNvParam', 33),
            mock.call(self.target, 'loadNvParam', 91),
            mock.call(self.target, 'loadNvParam', 78),
            mock.call(self.target, 'loadNvParam', 20),
            mock.call(self.target, 'loadNvParam', 21),
            mock.call(self.target, 'loadNvParam', 22),
            mock.call(self.target, 'loadNvParam', 23),
            mock.call(self.target, 'loadNvParam', 24),
            mock.call(self.target, 'loadNvParam', 25),
            mock.call(self.target, 'loadNvParam', 26),
            mock.call(self.target, 'loadNvParam', 27),
            mock.call(self.target, 'loadNvParam', 28),
            mock.call(self.target, 'loadNvParam', 29),
            mock.call(self.target, 'loadNvParam', 30),
            mock.call(self.target, 'loadNvParam', 31),
            mock.call(self.target, 'loadNvParam', 32)])

    @gen_test
    def test_vm_stat(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,27,32767,SM220'))

        yield node.get_vmstat(self.gus, self.target, 6, 3)

        self.gus.callback_ucast_rpc.assert_called_once_with(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2)

    @gen_test
    def test_vm_stat_retries_on_wrong_response(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((4, 42)),
            helpers.make_resolved_future((0x3005, None)),
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,27,32767,SM220'))
        ]

        yield node.get_vmstat(self.gus, self.target, 6, 3)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2),
            mock.call(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2),
            mock.call(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2),
        ])

    @gen_test
    def test_vm_stat_fails_on_enough_wrong_responses(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future((4, 42))

        with self.assertRaises(NoResponseError):
            yield node.get_vmstat(self.gus, self.target, 6, 3)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2),
            mock.call(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2),
            mock.call(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2),
        ])

    @gen_test
    def test_vm_stat_fails_on_no_responses(self):
        self.gus.callback_ucast_rpc.side_effect = NoResponseError()

        with self.assertRaises(NoResponseError):
            yield node.get_vmstat(self.gus, self.target, 6, 3)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2),
            mock.call(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2),
            mock.call(self.target, 'vmStat', 6, 3, callback_name='tellVmStat', retries=0, timeout=2),
        ])

    @gen_test
    def test_ping(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future((0x3005, None))

        data = yield node.ping(self.gus, self.target)

        self.assertEqual(data, {'lq': 0x30})
        self.gus.callback_ucast_rpc.assert_has_calls([mock.call(self.target, 'vmStat', 5, callback_name='tellVmStat', retries=0, timeout=2)])

    @gen_test
    def test_get_node_core_module(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,27,32767,SM220'))

        data = yield node.get_node_core_module(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_called_once_with(self.target, 'vmStat', 6, callback_name='tellVmStat', retries=0, timeout=2)
        self.assertEqual(data, {'core': '2.6.9', 'module': 'SM220UF1', 'platform': 'SM220'})


class NodeScanEnergyTestCase(AsyncTestCase, BaseTestCase):
    @gen_test
    def test_scan_energy(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future("ZZZZZZZZZZZZZZZZ")

        yield node.scan_energy(self.gus, self.target, num=3, delay=0)

        self.gus.callback_ucast_rpc.assert_has_calls([mock.call(self.target, 'scanEnergy'),
                                                      mock.call(self.target, 'scanEnergy')])

    def test_calculate_percent_clear(self):
        energies = {0: [0, 0, 0],
                    1: [3, 3, 3],
                    2: [0, 1, 2],
                    3: [1, 2, 3],
                    }
        cutoff = 2
        weights = node.calculate_percent_clear(energies, cutoff)
        self.assertAlmostEqual(weights["clear_percent_channel_00"], 0.00, places=3)
        self.assertAlmostEqual(weights["clear_percent_channel_01"], 1.00, places=3)
        self.assertAlmostEqual(weights["clear_percent_channel_02"], .3333, places=3)
        self.assertAlmostEqual(weights["clear_percent_channel_03"], .6667, places=3)

    @gen_test()
    def test_raw_scan_energies(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future("UVWXYZZZZZZZZZZZ")

        energies = yield node.raw_scan_energies(self.gus, self.target, num=3, delay=0)

        self.gus.callback_ucast_rpc.assert_has_calls([mock.call(self.target, 'scanEnergy'),
                                                      mock.call(self.target, 'scanEnergy')])

        expected_energies = {0: [85, 85, 85],
                             1: [86, 86, 86],
                             2: [87, 87, 87],
                             3: [88, 88, 88],
                             4: [89, 89, 89],
                             5: [90, 90, 90],
                             6: [90, 90, 90],
                             7: [90, 90, 90],
                             8: [90, 90, 90],
                             9: [90, 90, 90],
                             10: [90, 90, 90],
                             11: [90, 90, 90],
                             12: [90, 90, 90],
                             13: [90, 90, 90],
                             14: [90, 90, 90],
                             15: [90, 90, 90],
                             }

        self.assertDictEqual(energies, expected_energies)

    @gen_test()
    def test_percent_clear_energies(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future("WXYZ")  # WXYZ -> 87,88,89,90

        percent_clear = yield node.percent_clear_energies(self.gus, self.target, num=3, delay=0, cutoff=90)

        expected_percent_clear = {'clear_percent_channel_00': 0.0,
                                  'clear_percent_channel_01': 0.25,
                                  'clear_percent_channel_02': 0.5,
                                  'clear_percent_channel_03': 0.75,
                                  }

        self.assertDictEqual(percent_clear, expected_percent_clear)


class NodeOtaLoaderTestCase(AsyncTestCase, BaseTestCase):
    @gen_test
    def test_image_load(self):
        image = '.' * 100
        chunk_size = 93
        chunks = xrange(0, len(image), chunk_size)

        write_chunk_side_effects = [helpers.make_resolved_future((2, offset)) for offset in chunks]
        self.gus.callback_ucast_rpc.side_effect = [helpers.make_resolved_future((None,))] + write_chunk_side_effects

        yield node._ota_image_load(self.gus, self.target, image)

        write_chunk_calls = [mock.call(self.target, 'writeChunk', offset, image[offset:offset + chunk_size], callback_name='tellVmStat', retries=0) for offset in chunks]
        self.gus.callback_ucast_rpc.assert_has_calls([mock.call(self.target, 'eraseImage')] + write_chunk_calls + [mock.call(self.target, 'reboot')])

    @gen_test
    def test_image_load_failure_on_no_response(self):
        image = '.' * 100
        chunk_size = 93
        chunks = xrange(0, len(image), chunk_size)

        write_chunk_side_effects = [NoResponseError() for _ in chunks]
        self.gus.callback_ucast_rpc.side_effect = [helpers.make_resolved_future((None,))] + write_chunk_side_effects

        # The SPY filename is read, but its contents are not used, so this path could be any existing file
        yield node._ota_image_load(self.gus, self.target, image)

        write_chunk_retries = [mock.call(self.target, 'writeChunk', 0, image[0:0 + chunk_size], callback_name='tellVmStat', retries=0) for i in xrange(self.gus.network.default_cb_retries)]
        self.gus.callback_ucast_rpc.assert_has_calls([mock.call(self.target, 'eraseImage')] + write_chunk_retries)

    @gen_test
    def test_image_load_continue_through_echoed_response(self):
        image = '.' * 100
        chunk_size = 93
        chunks = xrange(0, len(image), chunk_size)

        write_chunk_side_effects = [helpers.make_resolved_future((2, offset)) for offset in chunks]
        # double each response
        with_echoes = list(chain(*zip(write_chunk_side_effects, write_chunk_side_effects)))
        self.gus.callback_ucast_rpc.side_effect = [helpers.make_resolved_future((None,))] + with_echoes

        # The SPY filename is read, but its contents are not used, so this path could be any existing file
        yield node._ota_image_load(self.gus, self.target, image)

        write_chunk_calls = [mock.call(self.target, 'writeChunk', offset, image[offset:offset + chunk_size], callback_name='tellVmStat', retries=0) for offset in chunks]
        calls_with_echoes = list(chain(*zip(write_chunk_calls, write_chunk_calls)))[1:]
        self.gus.callback_ucast_rpc.assert_has_calls([mock.call(self.target, 'eraseImage')] + calls_with_echoes + [mock.call(self.target, 'reboot')])

    @gen_test
    def test_image_load_fails_if_stuck_on_echoed_response(self):
        image = '.' * 100
        chunk_size = 93
        chunks = xrange(0, len(image), chunk_size)

        write_chunk_side_effects = [helpers.make_resolved_future((2, offset)) for offset in chunks]
        more_echoes = [write_chunk_side_effects[0]] * 10
        self.gus.callback_ucast_rpc.side_effect = [helpers.make_resolved_future((None,))] + more_echoes

        # The SPY filename is read, but its contents are not used, so this path could be any existing file
        with self.assertRaises(Exception):
            yield node._ota_image_load(self.gus, self.target, image)

    @gen_test
    def test_image_load_continue_through_single_bad_write(self):
        image = '.' * 100
        chunk_size = 93
        chunks = xrange(0, len(image), chunk_size)

        write_chunk_side_effects = [helpers.make_resolved_future((2, offset)) for offset in chunks]
        self.gus.callback_ucast_rpc.side_effect = [helpers.make_resolved_future((None,)), helpers.make_resolved_future((2, -9000))] + write_chunk_side_effects

        # The SPY filename is read, but its contents are not used, so this path could be any existing file
        yield node._ota_image_load(self.gus, self.target, image)

        write_chunk_calls = [mock.call(self.target, 'writeChunk', offset, image[offset:offset + chunk_size], callback_name='tellVmStat', retries=0) for offset in chunks]
        self.gus.callback_ucast_rpc.assert_has_calls([mock.call(self.target, 'eraseImage'), mock.call(self.target, 'writeChunk', 0, image[0:0 + chunk_size], callback_name='tellVmStat', retries=0)] + write_chunk_calls + [mock.call(self.target, 'reboot')])

    @gen_test
    def test_image_load_fails_with_many_bad_writes(self):
        image = '.' * 100
        chunk_size = 93
        chunks = xrange(0, len(image), chunk_size)

        write_chunk_side_effects = [helpers.make_resolved_future((2, int(offset / 2) + 3)) for offset in chunks]
        # repeat each response
        with_repeats = list(chain(*zip(write_chunk_side_effects, write_chunk_side_effects, write_chunk_side_effects, write_chunk_side_effects, write_chunk_side_effects)))
        self.gus.callback_ucast_rpc.side_effect = [helpers.make_resolved_future((None,))] + with_repeats

        # The SPY filename is read, but its contents are not used, so this path could be any existing file
        with self.assertRaises(Exception):
            yield node._ota_image_load(self.gus, self.target, image)


class NodeScriptTestCase(AsyncTestCase, BaseTestCase):

    def test_script_load_from_file(self):
        image = '.' * 100

        snaplib.ScriptsManager.getSnappyStringFromExport = mock.Mock(return_value=image)

        with mock.patch('snapstraction.node._ota_image_load') as imgload:
            node.ota_script_upload(self.gus, self.target, path.realpath(__file__))
            imgload.assert_called_once_with(self.gus, self.target, image, progress_cb=None, finished_cb=None)

    def test_script_load_from_image_bytes(self):
        image = '.' * 100

        with mock.patch('snapstraction.node._ota_image_load') as imgload:
            node.ota_script_upload(self.gus, self.target, image=image)
            imgload.assert_called_once_with(self.gus, self.target, image, progress_cb=None, finished_cb=None)

    @gen_test
    def test_script_erase(self):
        self.gus.callback_ucast_rpc.return_value = helpers.make_resolved_future((None,))

        yield node.script_erase(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_called_once_with(self.target, 'eraseImage')


class NodeFirmwareTestCase(AsyncTestCase, BaseTestCase):

    @gen_test
    def test_firmware_load_happy_path_from_firmware_wheel(self):
        import snap_firmware_2_6_9.metadata as md
        firmware_filename = node._get_file_path_for_version(md, 'SS200', '2.6.9')
        image = '.' * 100
        chunk_size = 93
        chunks = xrange(0, len(image), chunk_size)

        layout_mock = mock.Mock()
        layout_mock.getContents.return_value = image
        node._get_layout_from_filename = mock.Mock(return_value=layout_mock)

        write_chunk_side_effects = [helpers.make_resolved_future((2, offset)) for offset in chunks]
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,14,32767,SS200')),
            helpers.make_resolved_future((0xFFFF,)),
            helpers.make_resolved_future((None,)),
        ] + write_chunk_side_effects + [
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,14,32767,SS200'))
        ]

        data = yield node.ota_firmware_upload(self.gus, self.target, '2.6.9', None)

        self.assertEqual(data, {'core': '2.6.9', 'module': 'SS200', 'platform': 'SS200', 'firmware': firmware_filename})
        write_chunk_calls = [mock.call(self.target, 'writeChunk', offset, image[offset:offset + chunk_size], callback_name='tellVmStat', retries=0) for offset in chunks]

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'vmStat', 6, callback_name='tellVmStat', retries=0, timeout=2),
            mock.call(self.target, 'getInfo', 16),
            mock.call(self.target, 'eraseImage'),
        ] + write_chunk_calls + [mock.call(self.target, 'reboot')])

    @gen_test
    def test_firmware_load_over_unknown_current(self):
        firmware_filename = path.join('.', 'sfifiles', 'SS200_AES128_SnapV2.6.9.sfi')
        image = '.' * 100
        chunk_size = 93
        chunks = xrange(0, len(image), chunk_size)

        layout_mock = mock.Mock()
        layout_mock.getContents.return_value = image
        node._get_layout_from_filename = mock.Mock(return_value=layout_mock)

        write_chunk_side_effects = [helpers.make_resolved_future((2, offset)) for offset in chunks]
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((0xFFFF,)),
            helpers.make_resolved_future((None,)),
        ] + write_chunk_side_effects + [
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,14,32767,SS200'))
        ]

        data = yield node.ota_firmware_upload(self.gus, self.target, '2.6.9', firmware_filename)

        self.assertEqual(data, {'core': '2.6.9', 'module': '14', 'platform': 'SS200', 'firmware': path.abspath(firmware_filename)})
        write_chunk_calls = [mock.call(self.target, 'writeChunk', offset, image[offset:offset + chunk_size], callback_name='tellVmStat', retries=0) for offset in chunks]

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'getInfo', 16),
            mock.call(self.target, 'eraseImage'),
        ] + write_chunk_calls + [mock.call(self.target, 'reboot')])

    @gen_test
    def test_firmware_load_happy_file_override(self):
        firmware_filename = path.join('.', 'sfifiles', 'SS200_AES128_SnapV2.6.9.sfi')
        image = '.' * 100
        chunk_size = 93
        chunks = xrange(0, len(image), chunk_size)

        layout_mock = mock.Mock()
        layout_mock.getContents.return_value = image
        node._get_layout_from_filename = mock.Mock(return_value=layout_mock)

        write_chunk_side_effects = [helpers.make_resolved_future((2, offset)) for offset in chunks]
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((0xFFFF,)),
            helpers.make_resolved_future((None,)),
        ] + write_chunk_side_effects + [
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,14,32767,SS200'))
        ]

        data = yield node.ota_firmware_upload(self.gus, self.target, None, filename=firmware_filename)

        self.assertEqual(data, {'core': '2.6.9', 'module': '14', 'platform': 'SS200', 'firmware': path.abspath(firmware_filename)})
        write_chunk_calls = [mock.call(self.target, 'writeChunk', offset, image[offset:offset + chunk_size], callback_name='tellVmStat', retries=0) for offset in chunks]

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'getInfo', 16),
            mock.call(self.target, 'eraseImage'),
        ] + write_chunk_calls + [mock.call(self.target, 'reboot')])

    @gen_test
    def test_firmware_fails_on_module_not_installed(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,14,32767,SS200'))
        ]

        with self.assertRaises(Exception):
            yield node.ota_firmware_upload(self.gus, self.target, '7.11.0')

    @mock.patch('snapstraction.node.get_node_core_module')
    @gen_test
    def test_firmware_fails_on_file_when_module_not_in_filename(self, g):
        firmware_filename = path.join('.', 'sfifiles', 'SS200_AES128_SnapV2.6.9.sfi')

        g.side_effect = [
            helpers.make_resolved_future({'core': '2.6.9', 'module': 'Foo', 'platform': 'SS200'})
        ]
        with self.assertRaises(Exception):
            yield node.ota_firmware_upload(self.gus, self.target, '2.6.9', firmware_filename)

    @gen_test
    def test_firmware_fails_when_bank_space_doesnt_return(self):
        firmware_filename = path.join('.', 'sfifiles', 'SS200_AES128_SnapV2.6.9.sfi')

        image = '.' * 100
        layout_mock = mock.Mock()
        layout_mock.getContents.return_value = image
        node._get_layout_from_filename = mock.Mock(return_value=layout_mock)

        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future(None),
        ]

        with self.assertRaises(Exception) as e:
            yield node.ota_firmware_upload(self.gus, self.target, '2.6.9', firmware_filename)
            self.assertTrue("Couldn't get bank free space" in e.message)

    @gen_test
    def test_firmware_fails_when_bank_space_returns_zero(self):
        firmware_filename = path.join('.', 'sfifiles', 'SS200_AES128_SnapV2.6.9.sfi')

        image = '.' * 100
        layout_mock = mock.Mock()
        layout_mock.getContents.return_value = image
        node._get_layout_from_filename = mock.Mock(return_value=layout_mock)

        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((0,)),
        ]

        with self.assertRaises(Exception) as e:
            yield node.ota_firmware_upload(self.gus, self.target, '2.6.9', firmware_filename)
            self.assertTrue("doesn't support OTA Firmware Upgrades" in e.message)

    @mock.patch('snapstraction.node._ota_image_load')
    @mock.patch('snapstraction.node.sleep')
    @gen_test
    def test_firmware_fails_when_no_response_after_upload(self, l, s):
        firmware_filename = path.join('.', 'sfifiles', 'SS200_AES128_SnapV2.6.9.sfi')

        image = '.' * 100
        layout_mock = mock.Mock()
        layout_mock.getContents.return_value = image
        node._get_layout_from_filename = mock.Mock(return_value=layout_mock)

        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((0xFFFF,))
        ] + [NoResponseError] * 20

        l.side_effect = s.side_effect = [
            helpers.make_resolved_future(None)
        ]

        with mock.patch('snapstraction.node.timer') as mock_timer:
            mock_timer.side_effect = [1, 2, 200]

            with self.assertRaises(Exception) as e:
                yield node.ota_firmware_upload(self.gus, self.target, '2.6.9', firmware_filename)
                self.assertTrue("never responded" in e.message)

    @mock.patch('snapstraction.node._ota_image_load')
    @mock.patch('snapstraction.node.sleep')
    @gen_test
    def test_firmware_fails_if_firmware_doesnt_match_after_upload(self, l, s):
        firmware_filename = path.join('.', 'sfifiles', 'SS200_AES128_SnapV2.6.9.sfi')

        image = '.' * 100
        layout_mock = mock.Mock()
        layout_mock.getContents.return_value = image
        node._get_layout_from_filename = mock.Mock(return_value=layout_mock)

        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,14,32767,SS200')),
            helpers.make_resolved_future((0xFFFF,)),
            helpers.make_resolved_future((6, '2.7.1,(Dec  1 2015 / d267aae),1,1,5,14,32767,SS200'))
        ]

        l.side_effect = s.side_effect = [
            helpers.make_resolved_future(None)
        ]

        with self.assertRaises(Exception) as e:
            yield node.ota_firmware_upload(self.gus, self.target, '2.6.9', firmware_filename)
            self.assertTrue("did not complete successfully" in e.message)


class NodeStatsTestCase(AsyncTestCase, BaseTestCase):
    info_stats = [
        {'name': 'info.vendor', 'id': 0, 'value': 0, 'formatted_value': 'Synapse', 'func': 'getInfo'},
        {'name': 'info.primary-communications-interface', 'id': 1, 'value': 0, 'formatted_value': '802.15.4-based 2.4 GHz', 'func': 'getInfo'},
        {'name': 'info.cpu', 'id': 2, 'value': 5, 'formatted_value': 'ATmega', 'func': 'getInfo'},
        {'name': 'info.platform', 'id': 3, 'value': 17, 'formatted_value': 'Synapse SM200 Surface Mount Module', 'func': 'getInfo'},
        {'name': 'info.build', 'id': 4, 'value': 1, 'formatted_value': 'Release', 'func': 'getInfo'},
        {'name': 'info.version.major', 'id': 5, 'value': 2, 'func': 'getInfo'},
        {'name': 'info.version.minor', 'id': 6, 'value': 6, 'func': 'getInfo'},
        {'name': 'info.version.build', 'id': 7, 'value': 10, 'func': 'getInfo'},
        {'name': 'info.rpc.buffer', 'id': 9, 'value': 100, 'func': 'getInfo'},
        {'name': 'info.rpc.is-multicast', 'id': 10, 'value': 1, 'func': 'getInfo'},
        {'name': 'info.rpc.remaining-ttl', 'id': 11, 'value': 3, 'func': 'getInfo'},
        {'name': 'info.rpc.is-dmcast', 'id': 24, 'value': 0, 'func': 'getInfo'},
        {'name': 'info.mcast.groups', 'id': 27, 'value': 0x0001, 'func': 'getInfo'},
        {'name': 'info.dmcast.address-index', 'id': 26, 'value': 6, 'func': 'getInfo'},
        {'name': 'info.dmcast.original-ttl', 'id': 28, 'value': 3, 'func': 'getInfo'},
        {'name': 'info.strings.tiny-left', 'id': 19, 'value': 8, 'func': 'getInfo'},
        {'name': 'info.strings.small-left', 'id': 12, 'value': 8, 'func': 'getInfo'},
        {'name': 'info.strings.medium-left', 'id': 13, 'value': 8, 'func': 'getInfo'},
        {'name': 'info.strings.large-left', 'id': 20, 'value': 8, 'func': 'getInfo'},
        {'name': 'info.script.is-first-execution', 'id': 21, 'value': 0, 'func': 'getInfo'},
        {'name': 'info.script.base-address', 'id': 22, 'value': 0, 'func': 'getInfo'},
        {'name': 'info.script.base-bank', 'id': 23, 'value': 0, 'func': 'getInfo'},
        {'name': 'info.encryption-available', 'id': 8, 'value': 1, 'formatted_value': 'AES-128', 'func': 'getInfo'},
        {'name': 'info.route-table-size', 'id': 14, 'value': 8, 'func': 'getInfo'},
        {'name': 'info.route-table-entries', 'id': 15, 'value': 4, 'func': 'getInfo'},
        {'name': 'info.bank-free-space', 'id': 16, 'value': 0xF0FF - 0x10000, 'func': 'getInfo'},
        {'name': 'info.stdin-hook-trigger', 'id': 18, 'value': 0, 'func': 'getInfo'},
        {'name': 'info.read-and-reset-delay-factor', 'id': 25, 'value': 0, 'func': 'getInfo'},
    ]

    stat_stats = [
        {'name': 'stat.ds-null.tx-buffers', 'id': 0, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-uart0.rx-buffers', 'id': 1, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-uart0.tx-buffers', 'id': 2, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-uart1.rx-buffers', 'id': 3, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-uart1.tx-buffers', 'id': 4, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-transparent.rx-buffers', 'id': 5, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-transparent.tx-buffers', 'id': 6, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.packet-serial.rx-buffers', 'id': 7, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.packet-serial.tx-buffers', 'id': 8, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.radio.rx-buffers', 'id': 9, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.radio.tx-buffers', 'id': 10, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.radio.forwarded-unicasts', 'id': 11, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.packet-serial.forwarded-unicasts', 'id': 12, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.radio.forwarded-xcasts', 'id': 13, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.packet-serial.forwarded-xcasts', 'id': 14, 'value': 0x80, 'func': 'getStat'},
    ]

    all_stats = info_stats + stat_stats

    def verify_stats(self, stats, name):
        self.gus.callback_ucast_rpc.side_effect = [helpers.make_resolved_future((stat_info['value'],)) for stat_info in stats]

        data = yield node.get_stats(self.gus, self.target, name)

        self.assertEqual(data, {stat_info['name']: stat_info.get('formatted_value', stat_info['value']) for stat_info in stats})
        self.gus.callback_ucast_rpc.assert_has_calls([mock.call(self.target, stat_info['func'], stat_info['id']) for stat_info in stats])

    @gen_test
    def test_get_stats_info(self):
        # Test an arbitrary stat from info_stats
        stat = self.info_stats[0]
        self.verify_stats([stat], stat['name'])

    @gen_test
    def test_get_stats_stat(self):
        # Test an arbitrary stat from stat_stats
        stat = self.stat_stats[0]
        self.verify_stats([stat], stat['name'])

    @gen_test
    def test_get_stats_all_info(self):
        self.verify_stats(self.info_stats, 'info')

    @gen_test
    def test_get_stats_all_stats(self):
        self.verify_stats(self.stat_stats, 'stat')

    @gen_test
    def test_get_stats_all_everything(self):
        self.verify_stats(self.all_stats, 'all')


class PlatformLookupTests(unittest.TestCase):
    cases = [
        ((17, "SM200"), "SM200"),
        ((17, "RF200"), "SM200"),
        ((14, "RF200"), "RF200"),
    ]

    def test_cases(self):
        for (module_id, platform_str), expected in self.cases:
            self.assertEqual(expected, node.lookup_platform_by_id(module_id, platform_str))
