import unittest
import mock

import os
import tempfile

import snapstraction
from snapstraction.node_cache import NodeCache


class NodeCacheTests(unittest.TestCase):
    def setUp(self):
        fd, path = tempfile.mkstemp()
        self.path = path
        snapstraction.node_cache._env_override = mock.MagicMock(return_value=path)

    def test_cache_can_store(self):
        with NodeCache() as cache:
            cache.add('aabbcc')

        with NodeCache() as cache:
            self.assertEqual(len(cache), 1)
            self.assertIn('aabbcc', cache)

    def test_cache_can_remove_one(self):
        with NodeCache() as cache:
            cache.add('aabbcc')
            cache.add('ddeeff')

        with NodeCache() as cache:
            cache.discard('aabbcc')

        with NodeCache() as cache:
            self.assertNotIn('aabbcc', cache)
            self.assertIn('ddeeff', cache)
            self.assertEqual(len(cache), 1)

    def test_cache_can_remove_all(self):
        with NodeCache() as cache:
            cache.add('aabbcc')
            cache.add('ddeeff')

        with NodeCache() as cache:
            cache.clear()

        with NodeCache() as cache:
            self.assertEqual(len(cache), 0)
            self.assertNotIn('aabbcc', cache)

    def test_multiple_adds_merge(self):
        with NodeCache() as cache:
            cache.add('aabbcc')

        with NodeCache() as cache:
            cache.add('ddeeff')

        with NodeCache() as cache:
            self.assertIn('aabbcc', cache)
            self.assertIn('ddeeff', cache)
            self.assertEqual(len(cache), 2)

    def test_cache_can_iterate(self):
        with NodeCache() as cache:
            cache.add('aabbcc')
            cache.add('ddeeff')

        with NodeCache() as cache:
            self.assertEqual(len(cache), 2)
            self.assertEqual({item for item in cache}, cache.cache)
            for item in cache:
                self.assertIn(item, cache)

    def test_cache_is_not_updated_just_because_we_validated_an_address(self):
        with NodeCache() as cache:
            self.assertEqual(len(cache), 0)

        snapstraction.util.validate_snap_address('aabbcc')

        with NodeCache() as cache:
            self.assertEqual(len(cache), 0)
            self.assertNotIn('aabbcc', cache)

    def test_cache_is_not_updated_just_because_we_cleaned_up_an_address(self):
        with NodeCache() as cache:
            self.assertEqual(len(cache), 0)

        snapstraction.util.clean_up('aabbcc', '000001')

        with NodeCache() as cache:
            self.assertEqual(len(cache), 0)
            self.assertNotIn('aabbcc', cache)
            self.assertNotIn('000001', cache)

    def test_cache_is_not_updated_just_because_we_cleaned_up_a_bridge(self):
        with NodeCache() as cache:
            self.assertEqual(len(cache), 0)

        snapstraction.util.clean_up('bridge', '000001')

        with NodeCache() as cache:
            self.assertEqual(len(cache), 0)
            self.assertNotIn('aabbcc', cache)
            self.assertNotIn('000001', cache)


class BadNodeCacheTests(unittest.TestCase):
    def setUp(self):
        fd, path = tempfile.mkstemp()
        with open(path, 'wt') as f:
            f.write('9876\n')
            f.write('123456\n')
            f.write('# Not an address\n')
            f.write('\t\b\r\l\n')
            f.write('aabbcc\n')
        self.path = path
        snapstraction.node_cache._env_override = mock.MagicMock(return_value=path)

    def test_load_with_bad_addrs(self):
        with NodeCache() as cache:
            self.assertIn('123456', cache)
            self.assertIn('aabbcc', cache)
            self.assertEqual(len(cache), 2)

    def test_save_with_bad_addrs(self):
        with NodeCache() as cache:
            cache.clear()
            cache.add('123456')
            cache.add('\x12\x0000')
            cache.add('aabbcc')
            cache.add('')
            cache.add('\t\r\l\b')

        with NodeCache() as cache:
            self.assertIn('123456', cache)
            self.assertIn('aabbcc', cache)
            self.assertEqual(len(cache), 2)


class FileDoesNotExistTests(unittest.TestCase):
    def setUp(self):
        cache_path = os.path.join(os.path.dirname(__file__), 'DoesNotExist.cache')
        if os.path.exists(cache_path):
            os.unlink(cache_path)
        snapstraction.node_cache._env_override = mock.MagicMock(return_value=cache_path)

    def test_works_without_pre_existing_cache(self):
        with NodeCache() as cache:
            cache.add('aabbcc')
            self.assertIn('aabbcc', cache)
