import mock
from tornado.testing import AsyncTestCase, gen_test
from tornado.concurrent import Future
import unittest

from snappy import SnapConstantsHelper
from snapstraction import formatters
from snapstraction import node
from snapstraction.gus import Gus
from tests import helpers


class MockGus(Gus):
    def __new__(cls):
        obj = object.__new__(Gus)
        obj.__init__()
        return obj


def mock_gus():
    gus = mock.MagicMock()
    gus.bridge_address = '000001'
    gus.address = '111110'

    gus.network.default_cb_timeout = 2
    gus.network.default_cb_retries = 3

    gus.ucast_rpc.return_value = Future()
    gus.ucast_rpc.return_value.set_result(True)

    gus.mcast_rpc.return_value = Future()
    gus.mcast_rpc.return_value.set_result(True)

    gus.dmcast_rpc.return_value = Future()
    gus.dmcast_rpc.return_value.set_result(True)

    gus.callback_ucast_rpc.return_value = Future()
    gus.callback_ucast_rpc.return_value.set_result(True)

    gus.callback_mcast_rpc.return_value = Future()
    gus.callback_mcast_rpc.return_value.set_result(True)

    gus.callback_dmcast_rpc.return_value = Future()
    gus.callback_dmcast_rpc.return_value.set_result(True)

    gus.ucast_data.return_value = Future()
    gus.ucast_data.return_value.set_result(True)

    gus.mcast_data.return_value = Future()
    gus.mcast_data.return_value.set_result(True)

    gus.callback_ucast_data.return_value = Future()
    gus.callback_ucast_data.return_value.set_result(True)

    gus.callback_mcast_data.return_value = Future()
    gus.callback_mcast_data.return_value.set_result(True)

    gus.traceroute.return_value = Future()
    gus.traceroute.return_value.set_result(True)

    gus._base = MockGus()

    gus.on_rpc = gus._base.on_rpc
    gus.on_any_rpc = gus._base.on_any_rpc
    gus.on_data = gus._base.on_data
    gus.on_traceroute = gus._base.on_traceroute

    return gus


class BaseTestCase(unittest.TestCase):

    def setUp(self):
        self.gus = mock_gus()

        self.target = 'aabbcc'  # The address of the node to interact with
        self.snap_address = '012345'  # The address of the snap belt node
        self.pkt_content = 'foobar'  # The payload of incoming data packets

        self.gus.address = self.snap_address


class NodeInfoTestCase(AsyncTestCase, BaseTestCase):

    @gen_test
    def test_query_value_with_data(self):
        data = {'_errors': 0}

        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future(("That's amazing",))
        ]

        yield node.query_value_with_data(self.gus, self.target, data, 'amazing_thing', 'getAmazingThing', 12)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'getAmazingThing', 12),
        ])

        self.assertEqual(data, {'_errors': 0, 'amazing_thing': "That's amazing"})

    @gen_test
    def test_query_value_with_data_no_response(self):
        data = {'_errors': 0}

        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future(None)
        ]

        yield node.query_value_with_data(self.gus, self.target, data, 'amazing_thing', 'getAmazingThing', 12)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'getAmazingThing', 12),
        ])

        self.assertEqual(data, {'_errors': 1, 'amazing_thing': None})

    @gen_test
    def test_query_value_with_data_None_response(self):
        data = {'_errors': 0}

        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((None,))
        ]

        yield node.query_value_with_data(self.gus, self.target, data, 'amazing_thing', 'getAmazingThing', 12)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'getAmazingThing', 12),
        ])

        self.assertEqual(data, {'_errors': 0, 'amazing_thing': None})

    @gen_test
    def test_query_value_with_data_formatter(self):
        data = {'_errors': 0}

        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((0x12AB,))
        ]

        yield node.query_value_with_data(self.gus, self.target, data, 'amazing_thing', 'getAmazingThing', 12, fmt=formatters.format_as_hex)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'getAmazingThing', 12),
        ])

        self.assertEqual(data, {'_errors': 0, 'amazing_thing': '0x12ab'})

    @gen_test
    def test_query_nvparams_with_data(self):
        data = {'_errors': 0}

        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((0x12AB,))
        ]

        yield node.query_nvparams_with_data(self.gus, self.target, data, {SnapConstantsHelper.NV_NETWORK_ID: 'network'})

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', SnapConstantsHelper.NV_NETWORK_ID),
        ])

        self.assertEqual(data, {'_errors': 0, 'network': '0x12ab'})

    @gen_test
    def test_get_device_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future(('\x00\x1c\x2c\x00\x00\x0d\xec\xaf',)),
            helpers.make_resolved_future(('device-name-a',)),
            helpers.make_resolved_future(('device-type-b',)),
            helpers.make_resolved_future((0x051f,)),
            helpers.make_resolved_future(('platform-c',)),
            helpers.make_resolved_future((0x4012,)),
        ]

        data = yield node.device_info(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 2),
            mock.call(self.target, 'loadNvParam', 8),
            mock.call(self.target, 'loadNvParam', 10),
            mock.call(self.target, 'loadNvParam', 11),
            mock.call(self.target, 'loadNvParam', 41),
            mock.call(self.target, 'loadNvParam', 64),
        ])

        self.assertEqual(data, {
            '_errors': 0,
            'device.address': '00:1c:2c:00:00:0d:ec:af',
            'device.name': 'device-name-a',
            'device.type': 'device-type-b',
            'device.feature-bits': '0x051f',
            'device.platform': 'platform-c',
            'device.vendor-config-bits': '0x4012'
        })

    @gen_test
    def test_get_firmware_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,27,32767,SM220')),
        ]

        data = yield node.firmware_info(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'vmStat', 6, callback_name='tellVmStat', retries=0, timeout=2),
        ])

        self.assertEqual(data, {
            '_errors': 0,
            'device.core': '2.6.9',
            'firmware.category-id': 27,
            'device.platform': 'SM220',
            'firmware.version': '2.6.9 with AES-128 encryption',
            'firmware.category': 'SM220/RF220UF1/SN220',
        })

    @gen_test
    def test_get_mcast_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((False,)),
            helpers.make_resolved_future((False,)),
            helpers.make_resolved_future((True,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((None,)),
        ]

        data = yield node.mcast_info(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 5),
            mock.call(self.target, 'loadNvParam', 6),
            mock.call(self.target, 'loadNvParam', 78),
            mock.call(self.target, 'loadNvParam', 16),
            mock.call(self.target, 'loadNvParam', 17),
            mock.call(self.target, 'loadNvParam', 18),
            mock.call(self.target, 'loadNvParam', 33),
            mock.call(self.target, 'loadNvParam', 91),
        ])

        self.assertEqual(data, {
            '_errors': 0,
            'mcast.processed-groups': '0x0001',
            'mcast.forwarded-groups': '0x0001',
            'mcast.serial-forwarded-groups': None,
            'mcast.carrier-sense': False,
            'mcast.collision-detect': False,
            'mcast.collision-avoidance': True,
            'mcast.cs-cd-threshold': None,
            'mcast.csma-settings': None,
        })

    @gen_test
    def test_get_mesh_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((60000,)),
            helpers.make_resolved_future((1000,)),
            helpers.make_resolved_future((5000,)),
            helpers.make_resolved_future((5000,)),
            helpers.make_resolved_future((10000,)),
            helpers.make_resolved_future((3,)),
            helpers.make_resolved_future((500,)),
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((5,)),
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((0,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((127,)),
        ]

        data = yield node.mesh_info(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 20),
            mock.call(self.target, 'loadNvParam', 21),
            mock.call(self.target, 'loadNvParam', 22),
            mock.call(self.target, 'loadNvParam', 23),
            mock.call(self.target, 'loadNvParam', 24),
            mock.call(self.target, 'loadNvParam', 25),
            mock.call(self.target, 'loadNvParam', 26),
            mock.call(self.target, 'loadNvParam', 27),
            mock.call(self.target, 'loadNvParam', 28),
            mock.call(self.target, 'loadNvParam', 29),
            mock.call(self.target, 'loadNvParam', 30),
            mock.call(self.target, 'loadNvParam', 31),
            mock.call(self.target, 'loadNvParam', 32),
        ])

        self.assertEqual(data, {
            '_errors': 0,
            'mesh.routing.max-timeout': 60000,
            'mesh.routing.min-timeout': 1000,
            'mesh.routing.new-timeout': 5000,
            'mesh.routing.used-timeout': 5000,
            'mesh.routing.delete-timeout': 10000,
            'mesh.routing.rreq-retries': 3,
            'mesh.routing.rreq-wait-timeout': 500,
            'mesh.routing.initial-hop-limit': 1,
            'mesh.routing.max-hop-limit': 5,
            'mesh.seq-number': 1,
            'mesh.override': False,
            'mesh.penalty-lq-threshold': 127,
            'mesh.rejection-lq-threshold': 127,
        })

    @gen_test
    def test_get_network_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((4,)),
            helpers.make_resolved_future((0x1c2c,)),
            helpers.make_resolved_future((8,)),
            helpers.make_resolved_future((127,)),
            helpers.make_resolved_future((None,)),
        ]

        data = yield node.network_info(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 4),
            mock.call(self.target, 'loadNvParam', 3),
            mock.call(self.target, 'loadNvParam', 19),
            mock.call(self.target, 'loadNvParam', 39),
            mock.call(self.target, 'loadNvParam', 90),
        ])

        self.assertEqual(data, {
            '_errors': 0,
            'net.channel': 4,
            'net.network-id': '0x1c2c',
            'net.radio-unicast-retries': 8,
            'net.radio-lq-threshold': 127,
            'net.default-radio-rate': None,
        })

    @gen_test
    def test_get_script_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future(("ScriptKitty",)),
            helpers.make_resolved_future((0x1c2c,)),
        ]

        data = yield node.script_info(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'imageName'),
            mock.call(self.target, 'loadNvParam', 40),
        ])

        self.assertEqual(data, {
            '_errors': 0,
            'device.script-name': 'ScriptKitty',
            'device.script-crc': '0x1c2c',
        })

    @gen_test
    def test_get_security_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((1,)),
        ]

        data = yield node.security_info(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 50),
            mock.call(self.target, 'loadNvParam', 52),
        ])

        self.assertEqual(data, {
            '_errors': 0,
            'security.enable-encryption': 'AES-128',
            'security.lockdown-bitmask': '0x0001'
        })

    stat_stats = [
        {'name': 'stat.ds-null.tx-buffers', 'id': 0, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-uart0.rx-buffers', 'id': 1, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-uart0.tx-buffers', 'id': 2, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-uart1.rx-buffers', 'id': 3, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-uart1.tx-buffers', 'id': 4, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-transparent.rx-buffers', 'id': 5, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.ds-transparent.tx-buffers', 'id': 6, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.packet-serial.rx-buffers', 'id': 7, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.packet-serial.tx-buffers', 'id': 8, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.radio.rx-buffers', 'id': 9, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.radio.tx-buffers', 'id': 10, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.radio.forwarded-unicasts', 'id': 11, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.packet-serial.forwarded-unicasts', 'id': 12, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.radio.forwarded-xcasts', 'id': 13, 'value': 0x80, 'func': 'getStat'},
        {'name': 'stat.packet-serial.forwarded-xcasts', 'id': 14, 'value': 0x80, 'func': 'getStat'},
    ]

    def verify_stats(self, stats, name):
        self.gus.callback_ucast_rpc.side_effect = [helpers.make_resolved_future((stat_info['value'],)) for stat_info in stats]

        data = yield node.get_stats(self.gus, self.target, name)

        self.assertEqual(data, {stat_info['name']: stat_info.get('formatted_value', stat_info['value']) for stat_info in stats})
        self.gus.callback_ucast_rpc.assert_has_calls([mock.call(self.target, stat_info['func'], stat_info['id']) for stat_info in stats])

    @gen_test
    def test_get_stats_info(self):
        self.verify_stats(self.stat_stats, 'stat')

    @gen_test
    def test_get_summary_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((4,)),
            helpers.make_resolved_future((0x1c2c,)),
            helpers.make_resolved_future(('\x00\x1c\x2c\x00\x00\x0d\xec\xaf',)),
            helpers.make_resolved_future(('device-name-a',)),
            helpers.make_resolved_future(('device-type-b',)),
            helpers.make_resolved_future((0x051f,)),
            helpers.make_resolved_future(('platform-c',)),
            helpers.make_resolved_future((0x4002,)),
            helpers.make_resolved_future(("ScriptKitty",)),
            helpers.make_resolved_future((0x1c2c,)),
            helpers.make_resolved_future((6, '2.6.9,(Dec  1 2015 / d267aae),1,1,5,27,32767,SM220')),
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((1,)),
        ]

        data = yield node.summary_info(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'getChannel'),
            mock.call(self.target, 'getNetId'),
            mock.call(self.target, 'loadNvParam', 2),
            mock.call(self.target, 'loadNvParam', 8),
            mock.call(self.target, 'loadNvParam', 10),
            mock.call(self.target, 'loadNvParam', 11),
            mock.call(self.target, 'loadNvParam', 41),
            mock.call(self.target, 'loadNvParam', 64),
            mock.call(self.target, 'imageName'),
            mock.call(self.target, 'loadNvParam', 40),
            mock.call(self.target, 'vmStat', 6, callback_name='tellVmStat', retries=0, timeout=2),
            mock.call(self.target, 'loadNvParam', 50),
            mock.call(self.target, 'loadNvParam', 52),
        ])

        self.assertEqual(data, {
            '_errors': 0,
            'current.channel': 4,
            'current.network-id': '0x1c2c',
            'device.address': '00:1c:2c:00:00:0d:ec:af',
            'device.core': '2.6.9',
            'device.feature-bits': '0x051f',
            'device.name': 'device-name-a',
            'device.platform': 'SM220',
            'device.script-crc': '0x1c2c',
            'device.script-name': 'ScriptKitty',
            'device.type': 'device-type-b',
            'device.vendor-config-bits': '0x4002',
            'firmware.category-id': 27,
            'firmware.version': '2.6.9 with AES-128 encryption',
            'firmware.category': 'SM220/RF220UF1/SN220',
            'security.enable-encryption': 'AES-128',
            'security.lockdown-bitmask': '0x0001',
        })

    @gen_test
    def test_get_uart_info(self):
        self.gus.callback_ucast_rpc.side_effect = [
            helpers.make_resolved_future((1,)),
            helpers.make_resolved_future((5,)),
            helpers.make_resolved_future((75,)),
            helpers.make_resolved_future((0,)),
            helpers.make_resolved_future((None,)),
            helpers.make_resolved_future((None,)),
        ]

        data = yield node.uart_info(self.gus, self.target)

        self.gus.callback_ucast_rpc.assert_has_calls([
            mock.call(self.target, 'loadNvParam', 12),
            mock.call(self.target, 'loadNvParam', 13),
            mock.call(self.target, 'loadNvParam', 14),
            mock.call(self.target, 'loadNvParam', 15),
            mock.call(self.target, 'loadNvParam', 80),
            mock.call(self.target, 'loadNvParam', 81),
        ])

        self.assertEqual(data, {
            '_errors': 0,
            'uart.default': 1,
            'uart.buffering-timeout': 5,
            'uart.buffering-threshold': 75,
            'uart.intercharacter-timeout': 0,
            'uart.uart0-default-baudrate': None,
            'uart.uart1-default-baudrate': None,
        })
