import logging
import mock
import unittest

import tempfile

import snap_profile

from tests import helpers

from tornado.testing import AsyncTestCase, gen_test

from serialwrapper.ComUtils import ComPortScanner, PROBE_VER_SNAP2, PROBE_VER_SNYF1
from serialwrapper.ComUtils import PROBE_RESULT_FOUND, PROBE_RESULT_ERROR, PROBE_RESULT_READ_ERROR

from snapstraction import probe

log = logging.getLogger(__name__)


class MockComPortScanner(object):

    def __init__(self, *args, **kwargs):
        log.info("Creating Mock Com Port Scanner: %r, %r", args, kwargs)
        self.probeVer = None
        self.update_callback = None
        self.complete_callback = None

    def start(self, probeVer, autoRunCallback):
        self.probeVer = probeVer
        self.complete_callback = autoRunCallback

    def registerResultCallback(self, updateCallback):
        self.update_callback = updateCallback

    def mockUpdate(self, result):
        if self.update_callback:
            self.update_callback(result)

    def mockComplete(self, results):
        if self.complete_callback:
            self.complete_callback(results)


class SerialProbeTests(unittest.TestCase):

    @mock.patch.object(probe, '_probe')
    def test_probe_probes_for_snap_and_snyf(self, cmd):
        cmd.return_value = helpers.make_resolved_future([])

        probe.probe_serial_nodes()

        cmd.assert_has_calls([mock.call(force=False, sniffer=False, dry_run=False, progress_cb=None, finished_cb=None),
                              mock.call(force=False, sniffer=True, dry_run=False, progress_cb=None, finished_cb=None)])

    @mock.patch.object(probe, '_probe')
    def test_probe_passes_thru_args(self, cmd):
        cmd.return_value = helpers.make_resolved_future([])

        probe.probe_serial_nodes(force=True, dry_run=True)

        cmd.assert_has_calls([mock.call(force=True, sniffer=False, dry_run=True, progress_cb=None, finished_cb=None),
                              mock.call(force=True, sniffer=True, dry_run=True, progress_cb=None, finished_cb=None)])


class ProbeTests(AsyncTestCase):

    def setUp(self):
        # Make our tests operate on a temp file
        fd, path = tempfile.mkstemp()
        self.path = path
        snap_profile.profile._env_override = mock.MagicMock(return_value=path)

        with open(path, 'wt') as f:
            f.write("""
[default]
device = /dev/snap1
kind = serial
address =
""")
        super(ProbeTests, self).setUp()

    @gen_test
    def test_probe_calls_scanner(self):
        scanner = MockComPortScanner()
        probe.get_com_port_scanner = lambda x: scanner

        f = probe._probe(force=False, sniffer=False, dry_run=False)
        scanner.mockComplete([])
        results = yield f

        self.assertEqual(results, [])

    @gen_test
    def test_scanner_found_posix_port(self):
        scanner = MockComPortScanner()
        probe.get_com_port_scanner = lambda x: scanner

        f = probe._probe(force=True, sniffer=False, dry_run=True)
        r = ComPortScanner.ProbeResult(0, '/dev/ttyUSB0', PROBE_VER_SNAP2, result=PROBE_RESULT_FOUND)
        r.portStr = r.port
        scanner.mockUpdate(r)
        scanner.mockComplete([r])
        results = yield f

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].kind, 'serial')
        self.assertEqual(results[0].device, '/dev/ttyUSB0')
        self.assertEqual(results[0].name, 'ttyUSB0')

    @gen_test
    def test_scanner_found_COM_port(self):
        scanner = MockComPortScanner()
        probe.get_com_port_scanner = lambda x: scanner

        f = probe._probe(force=True, sniffer=False, dry_run=True)
        r = ComPortScanner.ProbeResult(0, 'COM13', PROBE_VER_SNAP2, result=PROBE_RESULT_FOUND)
        r.portStr = r.port
        scanner.mockUpdate(r)
        scanner.mockComplete([r])
        results = yield f

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].kind, 'serial')
        self.assertEqual(results[0].device, 'COM13')
        self.assertEqual(results[0].name, 'COM13')

    @gen_test
    def test_scanner_found_old_paddleboard(self):
        scanner = MockComPortScanner()
        probe.get_com_port_scanner = lambda x: scanner

        f = probe._probe(force=True, sniffer=False, dry_run=True)
        # Handgenerating the Pyserial names here
        r = ComPortScanner.ProbeResult(2, 0, PROBE_VER_SNAP2, result=PROBE_RESULT_FOUND)
        r.portStr = "USB0"
        scanner.mockUpdate(r)
        scanner.mockComplete([r])
        results = yield f

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].kind, 'serial')
        self.assertEqual(results[0].device, 'USB0')
        self.assertEqual(results[0].name, 'USB0')

    @gen_test
    def test_scanner_found_ss200(self):
        scanner = MockComPortScanner()
        probe.get_com_port_scanner = lambda x: scanner

        f = probe._probe(force=True, sniffer=False, dry_run=True)
        # Handgenerating the Pyserial names here
        r = ComPortScanner.ProbeResult(5, 0, PROBE_VER_SNAP2, result=PROBE_RESULT_FOUND)
        r.portStr = "SNAPstick0"
        scanner.mockUpdate(r)
        scanner.mockComplete([r])
        results = yield f

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].kind, 'serial')
        self.assertEqual(results[0].device, 'SnapStick0')
        self.assertEqual(results[0].name, 'SNAPstick0')

    @gen_test
    def test_scanner_found_sniffer(self):
        scanner = MockComPortScanner()
        probe.get_com_port_scanner = lambda x: scanner

        f = probe._probe(force=True, sniffer=False, dry_run=True)
        # Handgenerating the Pyserial names here
        r = ComPortScanner.ProbeResult(5, 0, PROBE_VER_SNYF1, result=PROBE_RESULT_FOUND)
        r.portStr = "SNAPstick0"
        scanner.mockUpdate(r)
        scanner.mockComplete([r])
        results = yield f

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].kind, 'serial')
        self.assertEqual(results[0].device, 'SnapStick0')
        self.assertEqual(results[0].name, 'SNAPstick0')

    @gen_test
    def test_no_profile_for_bad_reads(self):
        scanner = MockComPortScanner()
        probe.get_com_port_scanner = lambda x: scanner

        f = probe._probe(force=False, sniffer=False, dry_run=True)
        # Handgenerating the Pyserial names here
        r = ComPortScanner.ProbeResult(5, 0, PROBE_VER_SNAP2, result=PROBE_RESULT_READ_ERROR)
        r.portStr = "SNAPstick0"
        scanner.mockUpdate(r)
        scanner.mockComplete([r])
        results = yield f

        self.assertEqual(results, [])

    @gen_test
    def test_no_profile_for_errors_even_with_force(self):
        scanner = MockComPortScanner()
        probe.get_com_port_scanner = lambda x: scanner

        f = probe._probe(force=True, sniffer=False, dry_run=True)
        # Handgenerating the Pyserial names here
        r = ComPortScanner.ProbeResult(5, 0, PROBE_VER_SNAP2, result=PROBE_RESULT_ERROR)
        r.portStr = "SNAPstick0"
        scanner.mockUpdate(r)
        scanner.mockComplete([r])
        results = yield f

        self.assertEqual(results, [])


class KnownDeviceProbeTests(AsyncTestCase):

    def setUp(self):
        # Make our tests operate on a temp file
        fd, path = tempfile.mkstemp()
        self.path = path
        snap_profile.profile._env_override = mock.MagicMock(return_value=path)

        with open(path, 'wt') as f:
            f.write("""
[default]
device = /dev/snap1
kind = serial
address =
[SNAPstick0]
device = SnapStick0
kind = serial
address =
""")
        super(KnownDeviceProbeTests, self).setUp()

    @gen_test
    def test_no_new_profile_for_known_device(self):
        scanner = MockComPortScanner()
        probe.get_com_port_scanner = lambda x: scanner

        f = probe._probe(force=True, sniffer=False, dry_run=True)
        # Handgenerating the Pyserial names here
        r = ComPortScanner.ProbeResult(5, 0, PROBE_VER_SNAP2, result=PROBE_RESULT_FOUND)
        r.portStr = "SNAPstick0"
        scanner.mockUpdate(r)
        scanner.mockComplete([r])
        results = yield f

        self.assertEqual(results, [])
