"""Unit tests for topology gathering"""

import binascii
import mock
import unittest

from textwrap import dedent

from tornado.gen import sleep
from tornado.testing import AsyncTestCase, gen_test

import snapstraction
from snapstraction import topology
from snapstraction.gus import NoResponseError

from tests import helpers
from .test_node import BaseTestCase


class ChunkStringTests(unittest.TestCase):
    """Unit tests for topology utility string chunk function"""

    def test_chunk_string_chunks_None(self):
        self.assertEqual([], topology._chunk_string(None, 1))

    def test_chunk_string_chunks_Empty(self):
        self.assertEqual([], topology._chunk_string("", 1))

    def test_chunk_string_chunks_too_large(self):
        self.assertEqual(["abc"], topology._chunk_string("abc", 4))

    def test_chunk_string_chunks_normal(self):
        self.assertEqual(["abcde", "fghij", "klmno"], topology._chunk_string("abcdefghijklmno", 5))

    def test_chunk_string_chunks_bad_size(self):
        self.assertEqual(["abc"], topology._chunk_string("abc", None))


class TopoParserTests(unittest.TestCase):
    """Unit tests for topology parsing function"""

    def test_parsing_None_src(self):
        self.assertEqual({}, topology._parse_topo_response(None, None))

    def test_parsing_None_data(self):
        self.assertEqual({}, topology._parse_topo_response(None, None))

    def test_parsing_Empty_src(self):
        self.assertEqual({}, topology._parse_topo_response("", None))

    def test_parsing_Empty_data(self):
        self.assertEqual({}, topology._parse_topo_response("ff0000", ""))

    def test_parsing_too_small(self):
        self.assertEqual({}, topology._parse_topo_response("ff0000", "abc"))

    def test_parsing_too_large(self):
        self.assertEqual({}, topology._parse_topo_response("ff0000", "abcdefghij"))

    def test_parsing_normal(self):
        self.assertEqual({("abc123", "0a1b2c"): 50,
                          ("0a1b2c", "abc123"): 40},
                         topology._parse_topo_response("abc123", "\x0a\x1b\x2c\x32\x28"))


class MergeTopoResultsTests(unittest.TestCase):
    """Unit tests for merging topology results"""

    def test_merging_None(self):
        self.assertEqual({}, topology._merge_topo_responses(None))

    def test_merging_Empty(self):
        self.assertEqual({}, topology._merge_topo_responses({}))

    def test_parsing_single(self):
        self.assertEqual({("abc123", "0a1b2c"): 50,
                          ("0a1b2c", "abc123"): 40},
                         topology._merge_topo_responses(
                             {("abc123", "0a1b2c"): 50,
                              ("0a1b2c", "abc123"): 40}))

    def test_parsing_multiple(self):
        self.assertEqual({("abc123", "0a1b2c"): 50,
                          ("0a1b2c", "abc123"): 40},
                         topology._merge_topo_responses(
                             {("abc123", "0a1b2c"): [20, 80, 50],
                              ("0a1b2c", "abc123"): [38, 42, 10, 70]}))


class NodeCensusTests(AsyncTestCase, BaseTestCase):
    """Unit tests for gathering census of a single node"""

    @gen_test
    def test_node_census(self):
        self.gus.network.default_cb_timeout = .5
        self.gus.network.default_cb_retries = 2

        self.gus.ucast_rpc.return_value = helpers.make_resolved_future((True,))

        topo_f = topology.take_node_census(self.gus, self.target)
        # Let the loop run, briefly
        yield sleep(0.1)
        # Fake out a response
        self.gus._base._on_rpc_recv(self.target, 'tellVmStat', 12 + (42 << 8), "\x0a\x1b\x2c\x32\x28")
        # Let it finish
        topo = yield topo_f

        # Since we got a response quickly, it should only send the request once
        self.gus.ucast_rpc.assert_has_calls([mock.call(self.target, 'vmStat', 12)])

        self.assertEqual({'root': self.target, 'topology': {('0a1b2c', 'aabbcc'): 40, ('aabbcc', '0a1b2c'): 50}}, topo)

    @gen_test
    def test_node_census_for_not_responding(self):
        self.gus.network.default_cb_timeout = .5
        self.gus.network.default_cb_retries = 2

        self.gus.ucast_rpc.return_value = helpers.make_resolved_future((False,))

        topo = yield topology.take_node_census(self.gus, self.target)

        self.gus.ucast_rpc.assert_has_calls([mock.call(self.target, 'vmStat', 12),
                                             mock.call(self.target, 'vmStat', 12)])

        self.assertEqual({'root': self.target, 'topology': {}}, topo)


class NetworkCensusTests(AsyncTestCase, BaseTestCase):
    """Unit tests for gathering census of a network"""

    @mock.patch.object(snapstraction.topology, 'take_node_census')
    @gen_test
    def test_network_census(self, node_census):
        node_census.side_effect = [
            helpers.make_resolved_future({'root': self.target, 'topology': {('0a1b2c', 'aabbcc'): 40, ('aabbcc', '0a1b2c'): 50}}),
            helpers.make_resolved_future({'root': '0a1b2c', 'topology': {('0a1b2c', 'aabbcc'): 50, ('aabbcc', '0a1b2c'): 40}}),
        ]

        topo = yield topology.take_network_census(self.gus, self.target)

        node_census.assert_has_calls([mock.call(self.gus, self.target),
                                      mock.call(self.gus, '0a1b2c')])

        self.assertEqual({'root': self.target, 'topology': {('0a1b2c', 'aabbcc'): 45, ('aabbcc', '0a1b2c'): 45}}, topo)

    @mock.patch.object(snapstraction.topology, 'take_node_census')
    @gen_test
    def test_network_census_partial_not_responding(self, node_census):
        node_census.side_effect = [
            helpers.make_resolved_future({'root': self.target, 'topology': {('0a1b2c', 'aabbcc'): 40, ('aabbcc', '0a1b2c'): 50}}),
            helpers.make_resolved_future({'root': '0a1b2c', 'topology': {}}),
        ]

        topo = yield topology.take_network_census(self.gus, self.target)

        node_census.assert_has_calls([mock.call(self.gus, self.target),
                                      mock.call(self.gus, '0a1b2c')])

        # Resultant topo should be based only on the data we actually got
        self.assertEqual({'root': self.target, 'topology': {('0a1b2c', 'aabbcc'): 40, ('aabbcc', '0a1b2c'): 50}}, topo)

    @mock.patch.object(snapstraction.topology, 'take_node_census')
    @gen_test
    def test_network_census_does_not_query_weak_links(self, node_census):
        node_census.side_effect = [
            helpers.make_resolved_future({'root': self.target, 'topology': {('0a1b2c', 'aabbcc'): 96, ('aabbcc', '0a1b2c'): 96}}),
        ]

        topo = yield topology.take_network_census(self.gus, self.target, max_lq=80)

        node_census.assert_called_once_with(self.gus, self.target)

        # Resultant topo should be based only on the data we actually got
        self.assertEqual({'root': self.target, 'topology': {('0a1b2c', 'aabbcc'): 96, ('aabbcc', '0a1b2c'): 96}}, topo)
