import unittest
import mock
import os
import sys
from testfixtures import LogCapture

from distutils.sysconfig import get_python_lib
from mock import patch
import logging

import snapstraction.util


class FirstOrNoneTests(unittest.TestCase):

    def test_non_tuple_result(self):
        self.assertEqual('foo', snapstraction.util.first_or_none('foo'))

    def test_tuple_result(self):
        self.assertEqual('foo', snapstraction.util.first_or_none(('foo', 'bar', 'baz')))

    def test_None_result(self):
        self.assertIsNone(snapstraction.util.first_or_none(None))

    def test_list_result(self):
        self.assertEqual('foo', snapstraction.util.first_or_none(['foo', 'bar', 'baz']))


class SNAPParsingTests(unittest.TestCase):

    def test_base_10_integer(self):
        self.assertEqual(10, snapstraction.util.parse_value_for_snap('10'))

    def test_base_16_integer(self):
        self.assertEqual(16, snapstraction.util.parse_value_for_snap('0x10'))

    def test_string_escape_string(self):
        self.assertEqual('\xde\xad\xbe\xef', snapstraction.util.parse_value_for_snap(r'\xde\xad\xbe\xef'))
        self.assertEqual('decafbad\xde\xca\xfb\xad', snapstraction.util.parse_value_for_snap(r'decafbad\xde\xca\xfb\xad'))

    def test_empty_string(self):
        self.assertEqual('', snapstraction.util.parse_value_for_snap(''))
        self.assertEqual('', snapstraction.util.parse_value_for_snap('""'))
        self.assertEqual('', snapstraction.util.parse_value_for_snap("''"))

    def test_string(self):
        self.assertEqual('foo', snapstraction.util.parse_value_for_snap('foo'))

    def test_string_with_quotes(self):
        self.assertEqual('foo', snapstraction.util.parse_value_for_snap('"foo"'))
        self.assertEqual('foo', snapstraction.util.parse_value_for_snap("'foo'"))

    def test_true_boolean(self):
        self.assertEqual(True, snapstraction.util.parse_value_for_snap('true'))
        self.assertEqual(True, snapstraction.util.parse_value_for_snap('True'))
        self.assertEqual(True, snapstraction.util.parse_value_for_snap('TRUE'))

    def test_false_boolean(self):
        self.assertEqual(False, snapstraction.util.parse_value_for_snap('false'))
        self.assertEqual(False, snapstraction.util.parse_value_for_snap('False'))
        self.assertEqual(False, snapstraction.util.parse_value_for_snap('FALSE'))

    def test_None(self):
        self.assertIsNone(snapstraction.util.parse_value_for_snap('none'))
        self.assertIsNone(snapstraction.util.parse_value_for_snap('None'))
        self.assertIsNone(snapstraction.util.parse_value_for_snap('NONE'))
        self.assertIsNone(snapstraction.util.parse_value_for_snap(None))


class AddressValidationTests(unittest.TestCase):

    def test_addr_gets_lowered(self):
        self.assertEqual('aabbcc', snapstraction.util.validate_snap_address('AABBCC'))
        self.assertEqual('01bb23', snapstraction.util.validate_snap_address('01bB23'))

    def test_bridge_is_ok(self):
        self.assertEqual('bridge', snapstraction.util.validate_snap_address('bridge'))
        self.assertEqual('bridge', snapstraction.util.validate_snap_address('Bridge'))
        self.assertEqual('bridge', snapstraction.util.validate_snap_address('BRIDGE'))

    def test_bad_address_is_none(self):
        self.assertIsNone(snapstraction.util.validate_snap_address('XYZZY'))
        self.assertIsNone(snapstraction.util.validate_snap_address(''))
        self.assertIsNone(snapstraction.util.validate_snap_address('1234567890'))
        self.assertIsNone(snapstraction.util.validate_snap_address(None))


class AddressCleanupTests(unittest.TestCase):

    def test_normal_is_passed_through(self):
        self.assertEqual('aabbcc', snapstraction.util.clean_up('aabbcc', '000000'))

    def test_bridge_is_replaced(self):
        self.assertEqual('000000', snapstraction.util.clean_up('bridge', '000000'))
        self.assertEqual('000000', snapstraction.util.clean_up('Bridge', '000000'))
        self.assertEqual('000000', snapstraction.util.clean_up('BRIDGE', '000000'))


class FixUpTargetTests(unittest.TestCase):

    def test_fixed_cleans_up_scalar(self):

        mock_gus = mock.MagicMock()
        mock_gus.bridge_address = '010203'

        @snapstraction.util.fix_up_target
        def some_gus_func(gus, target, *args, **kwargs):
            return target

        self.assertEqual('aabbcc', some_gus_func(mock_gus, 'AABBCC'))

    def test_fixed_cleans_up_bridge_scalar(self):

        mock_gus = mock.MagicMock()
        mock_gus.bridge_address = '010203'

        @snapstraction.util.fix_up_target
        def some_gus_func(gus, target, *args, **kwargs):
            return target

        self.assertEqual('010203', some_gus_func(mock_gus, 'bridge'))

    def test_fixed_cleans_up_list(self):

        mock_gus = mock.MagicMock()
        mock_gus.bridge_address = '010203'

        @snapstraction.util.fix_up_target
        def some_gus_func(gus, target, *args, **kwargs):
            return target

        self.assertEqual(['010203', 'aabbcc', 'ddeeff'], some_gus_func(mock_gus, ['bridge', 'aabbcc', 'DDeeFF']))

    def test_fixed_raises_if_bad_target(self):

        mock_gus = mock.MagicMock()
        mock_gus.bridge_address = '010203'

        @snapstraction.util.fix_up_target
        def some_gus_func(gus, target, *args, **kwargs):
            return target

        with self.assertRaises(Exception):
            some_gus_func(mock_gus, None, None)


class VirtualenvTests(unittest.TestCase):
    def setUp(self):
        if os.environ.has_key('VIRTUAL_ENV'):
            self.virtual_env = os.environ['VIRTUAL_ENV']
        else:
            self.virtual_env = None

    def tearDown(self):
        if self.virtual_env:
            os.environ['VIRTUAL_ENV'] = self.virtual_env
        else:
            if os.environ.has_key('VIRTUAL_ENV'):
                del os.environ['VIRTUAL_ENV']

    def test_get_virtualenv_path_when_virtual_env_is_defined(self):
        os.environ['VIRTUAL_ENV'] = 'test'
        path = os.path.abspath(get_python_lib(prefix=os.environ['VIRTUAL_ENV']))
        self.assertEqual(snapstraction.util.get_virtualenv_path(), path)

    def test_executed_in_virtualenv_when_executing_in_virtual_env(self):
        os.environ['VIRTUAL_ENV'] = 'test'
        self.assertEqual(snapstraction.util.executing_in_virtualenv(), True)

    def test_executed_in_virtualenv_when_not_executing_in_virtual_env(self):
        if os.environ.has_key('VIRTUAL_ENV'):
            del os.environ['VIRTUAL_ENV']
        self.assertEqual(snapstraction.util.executing_in_virtualenv(), False)


class EnsurePathContainsPythonLibTests(unittest.TestCase):
    def setUp(self):
        self.original_sys_path = list(sys.path)

    def tearDown(self):
        sys.path = self.original_sys_path

    def _remove_path_element_from_sys_path(self, path_element):
        sys.path = filter(lambda e: e != path_element, sys.path)

    @patch('snapstraction.util.executing_in_virtualenv', return_value=False)
    @patch('snapstraction.util.get_python_lib', return_value="python_lib")
    def test_not_executing_in_virtual_env_and_python_lib_already_in_path(self, *args):
        if "python_lib" not in sys.path:
            sys.path.append("python_lib")
        before = list(sys.path)
        snapstraction.util.ensure_path_contains_python_lib()
        self.assertEqual(before, sys.path)

    @patch('snapstraction.util.executing_in_virtualenv', return_value=False)
    @patch('snapstraction.util.get_python_lib', return_value="python_lib")
    def test_not_executing_in_virtual_env_and_python_lib_not_yet_in_path(self, *args):
        self._remove_path_element_from_sys_path("python_lib")
        snapstraction.util.ensure_path_contains_python_lib()
        self.assertEqual("python_lib", sys.path[-1])

    @patch('snapstraction.util.executing_in_virtualenv', return_value=True)
    @patch('snapstraction.util.get_virtualenv_path', return_value="virtualenv_path")
    @patch('snapstraction.util.get_python_lib', return_value="python_lib")
    def test_executing_in_virtual_env_and_both_already_in_path(self, *args):
        if "python_lib" not in sys.path:
            sys.path.append("python_lib")
        if "virtualenv_path" not in sys.path:
            sys.path.insert(1, "virtualenv_path")
        before = list(sys.path)
        snapstraction.util.ensure_path_contains_python_lib()
        self.assertEqual(before, sys.path)

    @patch('snapstraction.util.executing_in_virtualenv', return_value=True)
    @patch('snapstraction.util.get_virtualenv_path', return_value="virtualenv_path")
    @patch('snapstraction.util.get_python_lib', return_value="python_lib")
    def test_executing_in_virtual_env_and_neither_in_path(self, *args):
        self._remove_path_element_from_sys_path("python_lib")
        self._remove_path_element_from_sys_path("virtualenv_path")
        snapstraction.util.ensure_path_contains_python_lib()
        self.assertEqual("virtualenv_path", sys.path[1])
        self.assertEqual("python_lib", sys.path[-1])

    @patch('snapstraction.util.executing_in_virtualenv', return_value=True)
    @patch('snapstraction.util.get_virtualenv_path', return_value="virtualenv_path")
    @patch('snapstraction.util.get_python_lib', return_value="python_lib")
    def test_executing_in_virtual_env_and_virtualenv_path_already_in_path(self, *args):
        self._remove_path_element_from_sys_path("python_lib")
        if "virtualenv_path" not in sys.path:
            sys.path.insert(1, "virtualenv_path")
        before_len = len(sys.path)
        snapstraction.util.ensure_path_contains_python_lib()
        self.assertEqual(before_len + 1, len(sys.path))
        self.assertEqual("python_lib", sys.path[-1])

    @patch('snapstraction.util.executing_in_virtualenv', return_value=True)
    @patch('snapstraction.util.get_virtualenv_path', return_value="virtualenv_path")
    @patch('snapstraction.util.get_python_lib', return_value="python_lib")
    def test_executing_in_virtual_env_and_python_lib_already_in_path(self, *args):
        self._remove_path_element_from_sys_path("virtualenv_path")
        if "python_lib" not in sys.path:
            sys.path.append("python_lib")
        before_len = len(sys.path)
        snapstraction.util.ensure_path_contains_python_lib()
        self.assertEqual(before_len + 1, len(sys.path))
        self.assertEqual("virtualenv_path", sys.path[1])


class TestOutputFilter(unittest.TestCase):
    def setUp(self):
        self.log = logging.getLogger("test")

    def test_OutputFilter_init(self):
        ef = snapstraction.util.OutputFilter("test")
        self.assertEquals(ef.log, self.log)
        self.assertEquals(ef.name, "test")

    def test_OutputFilter_errors_if_logger_not_existing(self):
        with self.assertRaises(KeyError):
            with snapstraction.util.OutputFilter("Not_Exist"):
                pass

    def test_OutputFilter_context_manager(self):
        self.assertEquals(self.log.filters, [])
        with snapstraction.util.OutputFilter("test") as ef:
            self.assertIn(ef, self.log.filters)
        self.assertEquals(self.log.filters, [])

    def test_OutputFilter_exit_reraises_exceptions(self):
        with self.assertRaises(ValueError):
            with snapstraction.util.OutputFilter("test") as ef:
                self.assertIn(ef, self.log.filters)
                raise ValueError
        self.assertEquals(self.log.filters, [])

    def test_OutputFilter_converts_given_log_errors_to_debugs(self):
        with LogCapture() as l:
            with snapstraction.util.OutputFilter("test"):
                self.log.critical("I should change.")
                self.log.error("This should log as debug, too.")
                self.log.info("Debug again.")
                self.log.warning("And my axe.")
                self.log.debug("This shouldn't be modified.")
        l.check(
            ("test", "DEBUG", "I should change."),
            ("test", "DEBUG", "This should log as debug, too."),
            ("test", "DEBUG", "Debug again."),
            ("test", "DEBUG", "And my axe."),
            ("test", "DEBUG", "This shouldn't be modified."),
        )

    def test_OutputFilter_does_not_convert_logs_from_other_loggers(self):
        foo = logging.getLogger("foo")
        with LogCapture() as l:
            with snapstraction.util.OutputFilter("test"):
                foo.info("Don't touch me!")
                foo.critical("Coming through!")
                self.log.error("Busted.")
        l.check(
            ('foo', 'INFO', "Don't touch me!"),
            ('foo', 'CRITICAL', 'Coming through!'),
            ('test', 'DEBUG', 'Busted.'),
        )
